# -*- coding: utf-8 -*-

# module siunits

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module is loaded by the unit module and is not intended be be
imported directly.  

The most of the units here are from the SI Brochure, 9th edition.
"""

from numpy import pi
from .ummy import ummy
from .unit import Conversion,Unit,MFraction
from .prefixedunit import PrefixedUnit
from .offsetunit import OffsetUnit,OffsetConversion


with Unit._builtin():
    #SI Base units
    _m = PrefixedUnit('metre','m',additional_names=('meter',),add_symbol=True,
                      order=1,description='SI unit of length',
                      base_description='SI base unit of length')
    Unit.alias('micron','um')
    _cm = Unit.unit('cm')
    _cm.description = 'SI unit of length and CGS base unit'
    
    _kg = PrefixedUnit('gram','g',add_symbol=True,order=1,base_prefix='kilo',
                      description='SI unit of mass',
                      base_description='SI base unit of mass')
    Unit.unit('g').description = 'SI unit of mass and CGS base unit'
    
    _s = PrefixedUnit('second','s',add_symbol=True,order=3,
                      description='SI unit of time',
                      base_description='SI and CGS base unit of time')
    
    _A = PrefixedUnit('ampere','A',add_symbol=True,order=4,
                      description='SI unit of electrical current',
                      base_description='SI base unit of electrical current')
    
    _K = PrefixedUnit('kelvin','K',add_symbol=True,order=5,
                      description='SI unit of temperature',
                      base_description='SI base unit of temperature')
    
    _mol = PrefixedUnit('mole','mol',add_symbol=True,order=6,
                      description='SI unit for amount of substance',
                      base_description='SI base unit for amount of substance')
    
    _cd = PrefixedUnit('candela','cd',add_symbol=True,order=7,
                      description='SI unit of luminosity',
                      base_description='SI base unit of luminosity')
    
    _ipk = Unit('international prototype kilogram','m(K)',
                Conversion(_kg,ummy(1,1.2e-8)),add_symbol=True,
                html_symbol='<i>m</i>(&#x1d4a6;)',latex_symbol='m({\\mathcal {K}})',
                description='IPK, le grand k')
    Unit.alias('IPK',_ipk)
    
    #Coherent derived units in the SI with special names and symbols
    _rad = PrefixedUnit('radian','rad',Conversion(_m*_m**-1,1),add_symbol=True,order=0,
                        description='unit of angle, angular unit')
    _sr = PrefixedUnit('steradian','sr',Conversion(_m**2*_m**-2,1),add_symbol=True,order=0,
                       description='unit of solid angle, angular unit')
    _Hz = PrefixedUnit('hertz','Hz',Conversion(_s**-1,1),add_symbol=True,order=0,
                       description='frequency')
    _N = PrefixedUnit('newton','N',Conversion(_m*_kg*_s**-2,1),add_symbol=True,order=0,
                      description='SI derived unit of force')
    _Pa = PrefixedUnit('pascal','Pa',Conversion(_N*_m**-2,1),add_symbol=True,order=0,
                       description='SI derived unit of pressure')
    _J = PrefixedUnit('joule','J',Conversion(_N*_m,1),add_symbol=True,order=0,
                      description='SI derived unit of energy')
    _W = PrefixedUnit('watt','W',Conversion(_J*_s**-1,1),add_symbol=True,order=0,
                      description='SI derived unit of power')
    _C = PrefixedUnit('coulomb','C',Conversion(_s*_A,1),add_symbol=True,order=0,
                      description='SI derived unit of electrical charge')
    _V = PrefixedUnit('volt','V',Conversion(_W*_A**-1,1),add_symbol=True,order=0,
                      description='SI derived unit of electric potential, voltage')
    PrefixedUnit('farad','F',Conversion(_C*_V**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical capacitance')
    _ohm = PrefixedUnit('ohm','\u03A9',Conversion(_V*_A**-1,1),add_symbol=True,order=0,
                 html_symbol='&Omega;',latex_symbol=r'\Omega',ascii_symbol='ohm',
                 description='SI derived unit of electrical resistance')
    Unit.alias('Ohm',_ohm)
    _siemens = PrefixedUnit('siemens','S',Conversion(_A*_V**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical conductance, susceptance and admittance')
    Unit.alias('mho',_siemens)
    _Wb = PrefixedUnit('weber','Wb',Conversion(_V*_s,1),add_symbol=True,order=0,
                       description='SI derived unit of magnetic flux')
    _T = PrefixedUnit('tesla','T',Conversion(_Wb*_m**-2,1),add_symbol=True,order=0,
                      description='SI derived unit of magnetic flux density')
    PrefixedUnit('henry','H',Conversion(_Wb*_A**-1,1),add_symbol=True,order=0,
                 description='SI derived unit of electrical inductance')
    
    _degC = OffsetUnit('degree Celsius','\u00B0C',OffsetConversion(_K,273.15),
               latex_symbol='^{\\circ}C',ascii_symbol = 'degC',add_symbol=True,order=0,
               description='unit of temperature')
    Unit.alias('degreeC',_degC)
    Unit.alias('degree-C',_degC)
    Unit.alias('deg-C',_degC)
    
    _lm = PrefixedUnit('lumen','lm',Conversion(_cd*_sr,1),add_symbol=True,order=0,
                       description='SI derived unit for luminous flux')
    PrefixedUnit('lux','lx',Conversion(_lm*_m**-2,1),add_symbol=True,order=0,
                 description='SI derived unit for illuminance and luminous emittance')
    PrefixedUnit('becquerel','Bq',Conversion(_s**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for radioactivity')
    PrefixedUnit('gray','Gy',Conversion(_J*_kg**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for ionizing radiation')
    PrefixedUnit('sievert','Sv',Conversion(_J*_kg**-1,1),add_symbol=True,order=0,
                 description='SI derived unit for ionizing radiation dose')
    PrefixedUnit('katal','kat',Conversion(_mol*_s**-1,1),add_symbol=True,order=0,
                 description='SI unit for catalytic activity')
        
    #Non-SI units accepted for use with the International System of Units
    _min = Unit('minute','min',Conversion(_s,60),add_symbol=True,order=0,
                description='unit of time')
    _h = Unit('hour','h',Conversion(_min,60),add_symbol=True,order=0,
              description='unit of time')
    _d = Unit('day','d',Conversion(_h,24),add_symbol=True,order=0,
              description='unit of time')
    Unit.alias('D',_d)
    _deg = Unit('degree','\t\u00B0',Conversion(_rad,pi/180),add_symbol=True,order=0,
                 latex_symbol='^{\\circ}',ascii_symbol='deg',
                 description='unit of angle, angular unit')
    _arcmin = Unit('arcminute',"\t'",Conversion(_deg,MFraction(1,60)),add_symbol=True,order=0,
                   short_name='arcmin',description='unit of angle, angular unit')
    Unit('arcsecond','\t"',Conversion(_arcmin,MFraction(1,60)),add_symbol=True,order=0,
         short_name='arcsec',description='unit of angle, angular unit')
    Unit('hectare','ha',Conversion('hm**2',1),add_symbol=True,order=0,
         description='unit of area')
    PrefixedUnit('litre','L',Conversion('dm**3',1),add_symbol=True,order=0,
                 additional_names=('liter',),description='unit of volume')
    _tonne = PrefixedUnit('tonne','t',Conversion(_kg,1000),add_symbol=True,order=0,
                          prefixes=['centi','deci','deca','hecto','kilo','mega',
                                    'giga','tera','peta','exa','zetta','yotta'],
                          description='unit of mass')
    Unit.alias('metric ton',_tonne)
    Unit.alias('metric-ton',_tonne)


# -*- coding: utf-8 -*-

# module indexed

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
A base class for the Unit and GummyConstant classes.
"""

from importlib import import_module
from html import escape


class _Builtin:
    """
    See the Indexed._builtin() class methos for the use of this class
    """
    def __init__(self,_cls):
        self._cls = _cls
        
    def  __enter__(self):
        self._cls._open_lib = self._cls._builtin_lib
    def __exit__(self,type, value, traceback):
        self._cls._open_lib = self._cls._lib

class Indexed:

    # to speed loading of the module, only wait to create instances until they
    # are needed.  If a instance cannot be found the modules below are loaded in 
    # reverse order.
    #_builtins_to_import = []
    
    # define these in a subclass at the class level:
    # _builtin_lib = {}
    # _lib = {}
    # _open_lib = _builtin_lib
    
    _case_sensitive = True
    
    @classmethod
    def alias(cls,alias,inst):
        """
        Creates an alias that can be used to reference an instance.
        
        Parameters
        ----------
        alias:  `str`
            a string containing the new alias

        inst:  `str` or `Indexed` instance
            A string referencing the `Indexed` instance that will be assigned
            the alias or the `Indexed` instance its self.
        """
        if isinstance(alias,str):
            if not cls._case_sensitive and ' ' in alias:
                alias = alias.lower()
            alias = alias.strip()
        inst = cls.get(inst)
        cls._open_lib[alias] = inst
        inst._add_alias(alias)
        
    @staticmethod
    def load(library_name):
        import_module(library_name)
    
    @classmethod
    def get(cls,name,exception=True):
        if isinstance(name,cls):
            return name
        if isinstance(name,str):
            if not cls._case_sensitive and ' ' in name:
                name = name.lower()
            name = name.strip()
        c = cls._lib.get(name)
        if c is None:
            c = cls._builtin_lib.get(name)
            if c is None:
                while len(cls._builtins_to_import) > 0:
                    import_module(cls._builtins_to_import.pop(),cls.__module__)
                    c = cls._builtin_lib.get(name)
                    if c is not None:
                        break    
    
        if c is None and exception:
            cls._raise_not_found(name)
        return c
    
    @classmethod
    def _raise_not_found(cls,name):
        raise ValueError(cls.object_name + ' "' + str(name) + '" was not found')
    
    @classmethod
    def _builtin(cls):
        """
        Allows Indexed instances to be added to Indexed._builtin_lib rather
        than Indexed._lib using a "with" statement.
        
        Example
        -------
        
        >>> with Unit._builtin():
        ...     one = Unit('1','',add_symbol=False)
        
        """
        return _Builtin(cls)
    
    @staticmethod
    def format_latex(text):
        return text
    
    def __init__(self,name,symbol=None,short_name=None,add_symbol=False,
                 html_symbol=None,latex_symbol=None,ascii_symbol=None,
                 description=None,case_sensitive=True):
        
        tolower = False
        if isinstance(name,str):
            if not self._case_sensitive and ' ' in name:
                tolower = True
            name = name.strip()
        self.name = name
        self.description=description
        
        if symbol is None:
            self.symbol = name
        else:
            self.symbol = symbol
            
        if html_symbol is None:
            self.html_symbol = escape(symbol)
        else:
            self.html_symbol = html_symbol
        if latex_symbol is None:
            sb = symbol.strip()
            sc = {
                '&': r'\&',
                '%': r'\%',
                '$': r'\$',
                '#': r'\#',
                '_': r'\_',
                '{': r'\{',
                '}': r'\}',
                '~': r'\textasciitilde{}',
                '^': r'\wedge{}',
                '\\': r'\textbackslash{}',
                '<': r'\textless{}',
                '>': r'\textgreater{}',
                ' ': r' \ '
                }
            self.latex_symbol = self.format_latex(''.join(sc[c] if c in sc 
                                                          else c 
                                                          for c in sb))
        else:
            self.latex_symbol = self.format_latex(latex_symbol)
        if ascii_symbol is None:
            self.ascii_symbol = ''.join(i if ord(i) < 128 else '_' for i in symbol)
        else:
            self.ascii_symbol=ascii_symbol
            
        self._aliases = set()
        
        if tolower:
            self._open_lib[name.lower()] = self
        else:
            self._open_lib[name] = self
        if add_symbol:
            if symbol.strip() != name:
                self._open_lib[symbol.strip()] = self
                self._aliases.add(symbol.strip())
            if (ascii_symbol is not None and ascii_symbol.strip() != name and 
                   ascii_symbol.strip() != symbol.strip()):
                self._open_lib[ascii_symbol.strip()] = self
                self._aliases.add(ascii_symbol.strip())
        if short_name is not None:
            tolower = False
            if isinstance(short_name,str):
                if not self._case_sensitive and ' ' in short_name:
                    tolower = True
                short_name = short_name.strip()
            self.short_name = short_name
            if tolower:
                short_name = short_name.lower()
            self._open_lib[short_name] = self
            self._aliases.add(short_name)
        else:
            if add_symbol:
                if ascii_symbol is not None:
                    self.short_name = ascii_symbol
                else:
                    self.short_name = symbol
            else:
                self.short_name = name
        
    @property
    def aliases(self):
        """read-only
        
        Returns a set of the unshadowed aliases of the instannce.  To add aliases
        to the unit use the `Indexed.alias` static method.
        """
        return {a for a in self._aliases if (self.get(a,exception=False) is self and a != self.name)}
    
    @property
    def shadowed_aliases(self):
        """read-only
        
        Returns a set of the shadowed aliases of the instance.
        """
        sha = {a for a in self._aliases if self.get(a,exception=False) is not self}
        if self.get(self.name) is not self:
            sha.add(self.name)
        return sha
    
    def _add_alias(self,alias):
        self._aliases.add(alias)
        
    def tostring(self,fmt=None,strip=True,**kwds):
        """
        Returns a string containing the symbol for the instance in the format
        given by the keyword `fmt` which may be set to a string the values 
        'html', 'latex', 'ascii' or 'unicode'.  
        """
        if fmt is None or fmt == 'unicode':
            ret = self.symbol
        elif fmt is None:
            ret = self.symbol
        elif fmt == 'html':
            ret = self.html_symbol
        elif fmt == 'latex':
            ret = self.latex_symbol
        elif fmt == 'ascii':
            ret =  self.ascii_symbol
        else:
            raise ValueError('format ' + str(fmt) + ' is not recognized')
        if strip:
            ret = ret.strip()
        return ret

    


# -*- coding: utf-8 -*-

# module dfunc

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
Class Dfunc is an abstract base class inherited by ummy and immy to provide 
support for numpy ufunc functions.
"""

import numpy as np
from numbers import Real,Integral

def _f_darctan2(x1,x2):
    return (x2/(x1**2 + x2**2),-x1/(x1**2 + x2**2))

def _f_dlogaddexp(x1,x2):
    return (np.exp(x1)/(np.exp(x1) + np.exp(x2)),
            np.exp(x2)/(np.exp(x1) + np.exp(x2)))
    
def _f_dlogaddexp2(x1,x2):
    return (2**x1*np.log(2)/(2**x1 + 2**x2),
           2**x2*np.log(2)/(2**x1 + 2**x2))

def _f_heaviside(x,h0):
    if not isinstance(h0,Real):
        raise TypeError('h0 must be a real number')
    return x._apply(np.heaviside,0,x,h0)

def _f_around(x,n=0):
    if not isinstance(n,Integral):
        raise TypeError('n must be an int')
    return x._nprnd(lambda y: np.around(y,n))


def _f_add(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__add__(x2)
    else:
        return x2.__radd__(x1)
    
def _f_sub(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__sub__(x2)
    else:
        return x2.__rsub__(x1)

def _f_mul(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__mul__(x2)
    else:
        return x2.__rmul__(x1)
    
def _f_div(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__truediv__(x2)
    else:
        return x2.__rtruediv__(x1)
    
def _f_fdiv(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__floordiv__(x2)
    else:
        return x2.__rfloordiv__(x1)
    
def _f_pow(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__pow__(x2)
    else:
        return x2.__rpow__(x1)
    
def _f_mod(x1,x2):
    if isinstance(x1,Dfunc):
        return x1.__mod__(x2)
    else:
        return x2.__rmod__(x1)


ddict = {np.sin: np.cos,
         np.cos: lambda x: -np.sin(x),
         np.tan:  lambda x: 1/np.cos(x)**2,
         np.arctan2: _f_darctan2,
         np.arcsin: lambda x: 1/np.sqrt(1 - x**2),
         np.arccos: lambda x: -1/np.sqrt(1 + x**2),
         np.arctan: lambda x: 1/(1 + x**2),
         np.sinh:  np.cosh,
         np.cosh:  np.sinh,
         np.tanh:  lambda x: -np.tanh(x)**2 + 1,
         np.arcsinh:  lambda x: 1/np.sqrt(1 + x**2),
         np.arccosh:  lambda x: 1/np.sqrt(1 - x**2),
         np.arctanh:  lambda x: 1/(1 - x**2),
         np.exp:  np.exp,
         np.exp2:  lambda x: np.log(2)*2**x,
         np.expm1:  np.exp,
         np.log:  lambda x: 1/x,
         np.log2:  lambda x: 1/(x*np.log(2)),
         np.log10:  lambda x: 1/(x*np.log(10)),
         np.log1p:  lambda x: 1/(1+x),
         np.logaddexp:  _f_dlogaddexp,
         np.logaddexp2:  _f_dlogaddexp2,
         np.heaviside: lambda x: 0, 
         np.sign:  lambda x: 0, 
        }

fdict = {np.angle:  lambda x: x.angle(),
         np.around:  _f_around,
         np.heaviside:  _f_heaviside,
         np.absolute:  lambda x: abs(x),
         np.add:  _f_add,
         np.subtract:  _f_sub,
         np.negative:  lambda x: -x,
         np.multiply:  _f_mul,
         np.divide:  _f_div,
         np.true_divide:  _f_div,
         np.floor_divide:  _f_fdiv,
         np.reciprocal:  lambda x: 1/x,
         np.power:  _f_pow,
         np.mod:  _f_mod,
         np.remainder: _f_mod,
         np.divmod:  lambda x1,x2: (_f_fdiv(x1,x2),_f_mod(x1,x2)),
         np.modf:  lambda x: (x%1,x//1),
         np.sqrt:  lambda x: x**0.5,
         np.square:  lambda x: x**2,
         np.cbrt:  lambda x: x**(1/3),
         np.real:  lambda x: x.real,
         np.imag:  lambda x: x.imag,
         np.conj:  lambda x: x.conjugate(),
         np.rint:  lambda x: x._nprnd(np.rint),
         np.fix:  lambda x: x._nprnd(np.fix),
         np.floor:  lambda x: x._nprnd(np.floor),
         np.ceil:  lambda x: x._nprnd(np.ceil),
         np.trunc:  lambda x: x._nprnd(np.trunc),
         np.isnan:  lambda x: np.isnan(x.x),
         np.isinf:  lambda x: np.isinf(x.x),
         np.isfinite:  lambda x: np.isfinite(x.x),
         np.isneginf:  lambda x: np.isneginf(x.x),
         np.isposinf:  lambda x: np.isposinf(x.x)
        }
if np.lib.NumpyVersion(np.__version__) >= '1.25.0':
    fdict[np.round] =  _f_around
else:
    fdict[np.round_] =  _f_around


try_fconvert = True
def _call(f,*x):
    if try_fconvert:
        try:
            return f(*x)
        except:
            x = [a.tofloat() if hasattr(a,'tofloat') else float(a) for a in x]
            return f(*x)

    return f(*x)
   
        
def _broadcast(f,x):
    # used for the binary operations __add__, ...
    if isinstance(x,np.ndarray):
        bx = np.broadcast(x)
        ret = np.array([f(*b) for b in bx])
        if bx.shape == ():
            ret = ret.item()
        else:
            ret = ret.reshape(bx.shape)
        return ret
    return _call(f,x)


class Dfunc:
    """
    Class `Dfunc` is an abstract base class that provides some support for numpy
    broadcasting for functions and operators.  An inheriting class must implement
    the `_apply(self,function,derivative,*args)`, `_napply(self,function,*args)`,
    and `tofloat(self)` methods, as well as `_add(x)`, `_radd(x)`, `_sub(x)`, ...
    """
    
    def tofloat(self):
        # this should return a copy of self with the x and u properties 
        # converted to float values
        raise NotImplementedError()
        
    @classmethod
    def apply(cls,function,derivative,*args):
        """
        A classmethod that applies a function to one or more gummy or jummy 
        objects propagating the uncertainty.
        
        Parameters
        ----------
        function: `function`
              The the function to be applied. For `gummy.apply`, 'function'
              should take one or more float arguments and return a float value 
              or float array.  For `jummy.apply`, 'function' may also take and
              return complex values.

        derivative:  `function`
              The name of a second function that gives the derivatives
              with respect to the arguments of `function`.  `derivative` should
              take an equal number of arguments as `function`.  If `function`
              takes one argument `derivative` should return a float and if
              `function` takes more than one argument then `derivative` should
              return a tuple, list or array of floats that contains the derivatives
              with respect to each argument.  In the case of `jummy.apply`, the
              derivatives with respect to each argument may be real or complex
              values, in which case `function` is assumed to be holomorphic.  Or
              the derivative may be a 2 x 2 matrix of the form:

                              [[ du/dx, du/dy ],
                               [ dv/dx, dv/dy ]]

             where function(x + j*y) = u + j*v.

        *args:  `gummy`, `jummy`, or `float`
              One or more arguments to which `function` will be applied.  These
              arguments need not all be `Dfunc` objects; arguments  such as
              floats will be taken to be constants with no uncertainty.
              They may also be numpy ndarrays in which case the usual numpy
              broadcasting rules apply.
              
        Returns
        -------
        `gummy`, `jummy`:
            If none of the arguments are `gummy` or `jummy`
            then the return value is the same type as the return value of `function`.
            Otherwise `gummy.apply` returns a `gummy` and `jummy.apply` returns either a
            `gummy` or a `jummy` depending on whether `function` has a float or
            a complex return value.
            
        
        Examples
        --------
            
        >>> import numpy as np
        >>> x = gummy(0.678,u=0.077)
        >>> gummy.apply(np.sin,np.cos,x)
        0.627 +/- 0.060
        
        >>> x = gummy(1.22,u=0.44)
        >>> y = gummy(3.44,u=0.67)
        >>> def dhypot(x,y):
        ...     return (x1/sqrt(x1**2 + x2**2),x2/np.sqrt(x1**2 + x2**2))
        >>> gummy.apply(np.hypot,dhypot,x,y)
        3.65 +/- 0.65
        """
        
        bargs = np.broadcast(*args)
        if bargs.shape == ():
            return _call(lambda *x: cls._apply(function,derivative,*x), *args) 
        
        ret = np.array([_call(lambda *x: cls._apply(function,derivative,*x), *a) for a in bargs])
        ret = ret.reshape(bargs.shape)
        return ret
    
    @classmethod
    def napply(cls,function,*args):
        """
        gummy.napply(function, arg1, arg2, ...) and
        jummy.napply(function, arg1, arg2, ...)
        
        A classmethod that applies a function to one or more gummy or jummy 
        objects propagating the uncertainty.  This method is similar to apply 
        except that the derivatives are computed numerically so a derivative 
        function does not need to be supplied.
        
        Parameters
        ----------
        function: `function`
            The the function to be applied. For `gummy.apply`, 'function'
            should take one or more float arguments and return a float value
            r float array.  For `jummy.apply`, 'function' may also take and
            return complex values.

        *args:  `gummy`, `jummy`, or `float`
              One or more arguments to which `function` will be applied.  These
              arguments need not all be `Dfunc` objects; arguments  such as
              floats will be taken to be constants with no uncertainty.
              They may also be numpy ndarrays in which case the usual numpy
              broadcasting rules apply.

        Returns
        -------
        `gummy`, `jummy`:
            If none of the arguments are `gummy` or `jummy`
            then the return value is the same type as the return value of `function`.
            Otherwise `gummy.apply` returns a `gummy` and `jummy.apply` returns either a
            `gummy` or a `jummy` depending on whether `function` has a float or
            a complex return value.
            
        
        Examples
        --------
            
        >>> import numpy as np
        >>> x = gummy(0.678,u=0.077)
        >>> gummy.napply(np.sin,x)
        0.627 +/- 0.060
        
        >>> x = gummy(1.22,u=0.44)
        >>> y = gummy(3.44,u=0.67)
        >>> gummy.napply(np.hypot,x,y)
        3.65 +/- 0.65
        """
        
        bargs = np.broadcast(*args)
        if bargs.shape == ():
            return _call(lambda *x: cls._napply(function,*x), *args)
        
        ret = np.array([_call(lambda *x: cls._napply(function,*x), *a) for a in bargs])
        ret = ret.reshape(bargs.shape)
        return ret
    
    def _ufunc(self,ufunc,*args,**kwds):
        if any(isinstance(a,np.ndarray) for a in args):
            args = [np.array(a) if isinstance(a,Dfunc) else a for a in args]
        try:
            return _call(lambda *x: self._apply(ufunc,ddict[ufunc],*x), *args)
        except KeyError:
            try:
                return _call(fdict[ufunc],*args)
            except KeyError:
                return self.napply(ufunc,*args)
       
    def __array_ufunc__(self,ufunc,method,*args,**kwds):
        if method != '__call__':
            return None
        
        return self._ufunc(ufunc,*args,**kwds)
            
    def __array_function__(self,func,method,*args,**kwds):        
        return self._ufunc(func,*args,**kwds)



# -*- coding: utf-8 -*-

from .test_create import TestCreate
from .test_operations import TestOperations
from .test_ubreakdown import TestUbreakdown
from .test_complex import TestComplex
from .test_misc import TestMisc
from .test_gummy import TestGummy


# -*- coding: utf-8 -*-

import metrolopy as uc
import numpy as np
import warnings
import unittest

from metrolopy.tests.common import rand,display

class TestCreate(unittest.TestCase):
    def test_gummy_init(self,n=None,exception_on_warning=True,prnt=False,plot=False):
        """
        Test gummy creation.
        """
        try:
            from scipy.stats import norm, t
            
            uc.gummy.p_method = None
            
            if n is None:
                if prnt or plot:
                    n = 100
                else:
                    n = 10000
            
            units=['m','lb','m**2 s**3/kg**4','degF','cm','%','kg']
            uunits = {'%':100, 'ppm':1e6, 'ppb':1e9, 'ms/s':1000}
            
            with warnings.catch_warnings():
                if exception_on_warning:
                    warnings.simplefilter('error')
                else:
                    warnings.simplefilter('ignore')
            
                for i in range(n):
                    
                    if rand.randint(2):
                        unit = units[rand.randint(len(units))]
                    else:
                        unit = uc.one
                        
                    if rand.randint(2):
                        if unit == 'm':
                            if rand.randint(2):
                                uunit = 'mm'
                            elif rand.randint(2):
                                uunit = 'm'
                            else:
                                uunit = 'in'
                        else:
                            if unit in ['degF','dB(SPL)','Np']:
                                uunit = None
                            else:
                                uunit = list(uunits.keys())[rand.randint(len(uunits))]
                    else:
                        uunit = None
                    
                    bayesian = bool(rand.randint(2))
                    
                    if rand.randint(2):
                        dof = rand.randint(3,20)
                    else:
                        dof = float('inf')
                        
                    if rand.randint(2):
                        if rand.randint(2):
                            k = 4*rand.rand() + 0.1
                            p = None
                        else:
                            p = rand.rand()/2 + 0.49
                            k = 1
                    else:
                        k = 1
                        p = None
                    
                    if rand.randint(2):
                        pmethods = ['loc','level of confidence','cp','coverage probability','gauss','ccp','ccp','conservative coverage probability','chebyshev']
                        uc.gummy.p_method = pmethods[rand.randint(len(pmethods))]
                    else:
                        uc.gummy.p_method = None
                    
                    if unit == 'dB(SPL)':
                        if rand.randint(10) == 0:
                            x = rand.randint(-10,10)
                        else:
                            x = 100.0*(2*rand.rand() - 1)
                    elif unit == 'Np':
                        if rand.randint(10) == 0:
                            x = rand.randint(-10,10)
                        else:
                            x = 20.0*(2*rand.rand() - 1)
                    else:
                        if rand.randint(10) == 0:
                            x = rand.randint(-60000,60000)
                        else:
                            x = (2*rand.rand() - 1)*10.0**rand.randint(-10,10)
                    if rand.randint(20) == 0:
                        x = 0
                            
                    if uunit in ['%','ppm','ppb','ms/s'] and not (unit == '%' and uunit == '%'):
                        U = 1.0e3*rand.rand()
                        u = abs(x)*U/uunits[uunit]
                    elif unit == 'm' and uunit == 'mm':
                        if x == 0:
                            U = 1.0e3*rand.rand()
                        else:
                            U = abs(x)*1.0e3*rand.rand()
                        u = U/1000
                    elif unit == 'm' and uunit == 'in':
                        if x == 0:
                            U = 100*rand.rand()
                        else:
                            U = abs(x)*100*rand.rand()
                        u = U*0.0254 
                    else:
                        if x == 0:
                            U = 1.0e12*rand.rand()
                        else:
                            if rand.randint(20) == 0:
                                U = 1e6*abs(x)*rand.rand()
                            elif rand.randint(20) == 0:
                                U = rand.randint(20)+1
                            else:
                                U = 1.1*abs(x)*rand.rand()
                        u = U
                            
                    if rand.randint(2):
                        name = chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65)
                    else:
                        name = None
                        
                    if rand.randint(2):
                        utype = chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65) + chr(rand.randint(20) + 65)
                    else:
                        utype = None
                        
                    if rand.randint(20) == 0:
                        self.assertTrue(uc.gummy(x) == x)
                        continue
                    
                    if rand.randint(20) == 0:
                        const = True
                        U = 0
                        u = 0
                    else:
                        if u == 0:
                            const = True
                        else:
                            const = False
                        
                    uc.gummy.bayesian = bayesian
                    if uunit is not None and rand.randint(2):
                        uu = uc.gummy(U,unit=uunit)
                        g = uc.gummy(x,u=uu,unit=unit,dof=dof,k=k,p=p,name=name,
                                     utype=utype)
                    else:
                        g = uc.gummy(x,u=U,unit=unit,dof=dof,k=k,p=p,uunit=uunit,
                                     name=name,utype=utype)
                        
                    if not const:
                        u = u/g.k
                        
                    orig_g = g
        
                    if rand.randint(10) == 0:
                        n = rand.randint(12)
                        #if bayesian and n > 0:
                           # dof = float('inf')
                        if n == 0:
                            nm = g.name
                            g = uc.gummy(g)
                            self.assertTrue(g.name is nm)
                            g.name = None
                        elif n == 1:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                g = 1*g
                        elif n == 2:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                g = g*1
                        elif n == 3:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                g = g/1
                        elif  n == 4:
                            if unit == 'degF':
                                v = uc.gummy(0,unit='degF-i')
                            elif unit is not uc.one or rand.randint(2):
                                v = uc.gummy(0,unit=unit)
                            else:
                                v = 0
                            g = g + v
                        elif n == 5:
                            if unit == 'degF':
                                v = uc.gummy(0,unit='degF-i')
                            elif unit is not uc.one or rand.randint(2):
                                v = uc.gummy(0,unit=unit)
                            else:
                                v = 0
                            g = v + g
                        elif n == 6:
                            if unit == 'degF':
                                v = uc.gummy(0,unit='degF-i')
                            elif unit is not uc.one or rand.randint(2):
                                v = uc.gummy(0,unit=unit)
                            else:
                                v = 0
                            g = g - v
                        elif n == 7:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            elif unit in ['dB(SPL)','Np']:
                                g =  g + uc.gummy(0,unit=unit)
                            else:
                                g = g**1
                        elif n == 8:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                g = +g
                        elif n == 9:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                g = 2.1*g
                                g = g/2.1
                        elif n == 10:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                if unit is not uc.one or rand.randint(2):
                                    v = uc.gummy(3.5*x,u=0.45*u,unit=unit)
                                else:
                                    v = 3.3*x
                                g = v - g
                                g = g - v
                                g = -g
                        else:
                            if unit == 'degF':
                                g =  g + uc.gummy(0,unit='degF-i')
                            else:
                                if x >= 0:
                                    g = abs(g)
                                else:
                                    g = -abs(g)
            
                        k = 1
                        p = None
                        reinit = True
                        uunit = None
                        name = None
                    else:
                        reinit = False
            
                    if x != 0 and g.x != 0:
                        self.assertTrue(abs(g.x - x)/abs(x) < 1e-10)
                    else:
                        self.assertTrue(abs(g.x - x) < 1e-10)
                        
                    if x != 0 and g.x != 0:
                        self.assertTrue(abs(g.x - x)/abs(x) < 1e-10)
                    else:
                        self.assertTrue(abs(g.x - x) < 1e-10)
                                    
                    if unit == 'degF':
                        self.assertTrue(g.unit in [uc.Unit.unit('degF'),uc.Unit.unit('degF-i')])
                    else:
                        self.assertTrue(g.unit is uc.Unit.unit(unit))
                    self.assertTrue(g.name == name)
                    
                    if const:
                        self.assertTrue(g.u == 0)
                        if unit is uc.one:
                            if x == 0:
                                self.assertTrue(abs(g - x) < 1e-10)
                            else:
                                self.assertTrue(abs(g - x)/x < 1e-10)
                        elif unit == '%':
                            if x == 0:
                                 self.assertTrue(abs(g - x/100) < 1e-10)
                            else:
                                self.assertTrue(abs((g - x/100)/(x/100)) < 1e-10)
                        elif unit == 'Np':
                            self.assertTrue(abs(g.convert(1) - np.exp(x))/np.exp(x) < 1e-10)
                        continue
                    
                    if uunit in ['%','ppm','ppb','dB','ms/s'] and unit != uunit:
                        self.assertTrue(g.uunit_is_rel)
                    else:
                        self.assertTrue(not g.uunit_is_rel)
                        
                    self.assertTrue((g.x - x)/u < 1e-6)
                    
                    if not uc.gummy.bayesian:
                        if dof == float('inf'):
                            self.assertTrue(g.dof == float('inf'))
                        else:
                            self.assertTrue((g.dof - dof)/dof < 1e-6)
                    
                    if uunit is not None and uunit != unit:
                        self.assertTrue(g.uunit is uc.Unit.unit(uunit))
                        
                    if g.p == 0:
                        self.assertTrue(p is None)
                        self.assertTrue(k <= 3)
                    else:
                        if p is None:
                            self.assertTrue(g.k == k)
                        else:
                            self.assertTrue(g.p == p)
                        
                        if uc.gummy.p_method in ['loc','level of confidence',None]:
                            if dof == float('inf'):
                                self.assertTrue(abs(g.k - norm.ppf(0.5 + g.p/2)) < 1e-6)
                            else:
                                if bayesian:
                                    if g.p is not None:
                                        self.assertTrue(abs(g.k - np.sqrt((dof-2)/dof)*t.ppf(0.5 + g.p/2,dof)) < 1e-6)
                                else:
                                    self.assertTrue(abs(g.k - t.ppf(0.5 + g.p/2,dof)) < 1e-6)
                                    
                        elif uc.gummy.p_method in ['cp','coverage probability','gauss']:
                            if dof == float('inf') or bayesian:
                                if g.p is not None:
                                    self.assertTrue(abs(g.k - 2/(3*(np.sqrt(1 - g.p)))) < 1e-6)
                            else:
                                if g.p is not None:
                                    self.assertTrue(abs(g.k - np.sqrt(dof/(dof-2))*(2/(3*(np.sqrt(1 - g.p))))) < 1e-6)                 
                        else:
                            if dof == float('inf') or bayesian:
                                if g.p is not None:
                                    self.assertTrue(abs(g.k - 1/np.sqrt(1 - g.p)) < 10.0**-6)
                            else:
                                if g.p is not None:
                                    self.assertTrue(abs(g.k - np.sqrt(dof/(dof-2))*(1/np.sqrt(1 - g.p))) < 1e-6)
                            
                    self.assertTrue(abs((g.u - u)/u) < 1e-10)
                    
                    if not reinit:
                        self.assertTrue(abs(g.U - U)/U < 1e-10)
                    
                    self.assertTrue(abs((g.u - u)/u) < 1e-10)
                    
                    self.assertTrue('?' not in g.tostring())
                    
                    if prnt:
                        print('------')
                        print()
                        if rand.randint(2):
                            styles = ['pm','pmi','concise','ueq','x','xf','u','uf','xunit','uunit']
                            g.style = styles[rand.randint(10)]
                            
                            g.show_k = bool(rand.randint(2))
                            g.show_dof = bool(rand.randint(2))
                            g.show_p = bool(rand.randint(2))
                            g.show_name = bool(rand.randint(2))
                            g.solidus = bool(rand.randint(2))
                            g.mulsep = bool(rand.randint(2))
                            if not rand.randint(5):
                                g.nsig = rand.randint(5) + 1
                            
                            self.assertTrue('?' not in g.tostring())
                            print(g.style,', show_k=',g.show_k,', show_dof=',g.show_dof,', show_p=',g.show_p,', show_name=',g.show_name,', solidus=',g.solidus,', mulsep=',g.mulsep,', nsig=',g.nsig,':')
                            print()
                        display(g)
                        print()
                            
                    if plot:
                        print('------')
                        print()
                        print(g)
                        g.sim()
                        if rand.randint(2):
                            styles = ['pmsim','pmsimi','mcisim','cisim','usim','ufsim']
                            g.style = styles[rand.randint(6)]
                        g.slashaxis = bool(rand.randint(2))
                        print(g.style,', slashaxis=',g.slashaxis,':')
                        print()
                        display(g)
                        print()
                        g.hist()
                        print()
                        
                    uc.gummy.bayesian = False
        finally:
            uc.gummy.p_method = 'loc'


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

# module constant

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
constant
"""

from .indexed import Indexed
from .gummy import gummy,jummy
from .unitutils import _mrtxt
from importlib import import_module
from .printing import PrettyPrinter
from .exceptions import ConstantNotFoundError
from .printing import _latex_math

class GummyConstant(gummy,Indexed):
    
    _case_sensitive = False
    return_toummy = False
    return_splonk = False
    
    _builtins_to_import = ['..builtin_constants']
    
    _builtin_lib = {}
    _lib = {}
    _open_lib = _lib
    
    @classmethod
    def _raise_not_found(cls,name):
        raise ConstantNotFoundError('constant "' + str(name) + '" was not found')
    
    def __new__(cls,x,u=0,unit=1,dof=float('inf'),k=1,p=None,uunit=None,
                 utype=None,name=None,symbol=None,short_name=None,
                 add_symbol=False, html_symbol=None,latex_symbol=None,
                 ascii_symbol=None,description=None,_key=None):
        if symbol is None or name is None:
            ret = gummy.__new__(gummy,x,u=u,unit=unit,dof=dof,k=k,p=p,
                                uunit=uunit,utype=utype,name=name)
            ret.__init__(x,u=u,unit=unit,dof=dof,k=k,p=p,uunit=uunit,
                         utype=utype,name=name)
            return ret
        
        return super().__new__(cls)
    
    def __init__(self,x,u=0,unit=1,dof=float('inf'),k=1,p=None,uunit=None,
                 utype=None,name=None,symbol=None,short_name=None,
                 add_symbol=False, html_symbol=None,latex_symbol=None,
                 ascii_symbol=None,description=None,_key=None):
        """
        A subclass of `gummy` that can be retrieved by name or alias from the
        constant library with the `constant` function.
        
        `GummyConstant` takes all the parameters that `gummy` takes, though 
        the `name` parameter is now required.  In addition it takes a required
        `symobol` parameter and several other optional parameters:
            
        Parameters
        ----------
        symbol:  `str`, required
            A unicode symbol.  If the `add_symbol` parameter is set to `True`, 
            then this symbol can also be used to access the unit with the 
            `constant` function.
            
        short_name: `str` or `None`
            a short name which can be used as an additional alias for the 
            constant in the constant library
        
        add_symbol: `bool`, optional
            If this is `True`, then the symbol can be used to look up the constant
            in the constant library.  The default is `False`
            
        html_symbol, latex_symbol, ascii_symbol:  `str` or `Mone`, optional
            html, latex, and ascii versions of the symbol if they are different
            from the unicode representation of the symbol.
            
        description:  `str` or `None`, optional
            a description of the constant
        """
        if name is None:
            name = symbol
            
        if html_symbol is None:
            html_symbol = '<i>' + symbol + '</i>'
            
        gummy.__init__(self,x,u=u,unit=unit,dof=dof,k=k,p=p,uunit=uunit,
                      utype=utype,name=name)
        Indexed.__init__(self,name,symbol=symbol,short_name=short_name,
                         add_symbol=add_symbol,html_symbol=html_symbol,
                         latex_symbol=latex_symbol,ascii_symbol=ascii_symbol,
                         description=description)
        
        self._key = _key
    
    def tostring(self,fmt=None,style=None,k=None,p=None,show_k=None,
                 show_p=None,show_dof=None,show_name=None,name=None,
                 norm=None,raw=False,nsig=None,solidus=None,
                 mulsep=None,**kwds):
                     
        if name is None:
            name = Indexed.tostring(self,fmt)
        
        return super().tostring(fmt=fmt,style=style,k=k,p=p,show_k=show_k,
                                show_p=show_p,show_dof=show_dof,
                                show_name=show_name,name=name,norm=norm,
                                raw=raw,nsig=nsig,solidus=solidus,
                                mulsep=mulsep)
    
    @property
    def unit(self):
        return self._unit
    @unit.setter
    def unit(self,u):
        name = self.name
        gummy.unit.fset(self,u)
        self.name = name
    
class JummyConstant(jummy,Indexed):
    _builtins_to_import = ['..builtin_constants']
    
    _builtin_lib = {}
    _lib = {}
    _open_lib = _lib
    
    @classmethod
    def _raise_not_found(cls,name):
        raise ConstantNotFoundError('constant "' + str(name) + '" was not found')
    
    def __new__(cls,real=None,imag=None,r=None,phi=None,cov=None,unit=1,
                 name=None,symbol=None,short_name=None,add_symbol=False,
                 html_symbol=None,latex_symbol=None,ascii_symbol=None,
                 description=None):
        if symbol is None:
            ret = jummy.__new__(jummy,real=real,imag=imag,r=r,phi=phi,cov=cov,
                                unit=unit,name=name)
            ret.__init__(real=real,imag=imag,r=r,phi=phi,cov=cov,unit=unit,
                         name=name)
            return ret
        
        return super().__new__(cls)
    
    def __init__(self,real=None,imag=None,r=None,phi=None,cov=None,unit=1,
                 name=None,symbol=None,short_name=None,add_symbol=False,
                 html_symbol=None,latex_symbol=None,ascii_symbol=None,
                 description=None):
        if name is None:
            name = symbol
            
        if html_symbol is None:
            html_symbol = '<i>' + symbol + '</i>'
            
        jummy.__init__(self,real=real,imag=imag,r=r,phi=phi,cov=cov,unit=1,
                       name=name)
        Indexed.__init__(self,name,symbol=symbol,short_name=short_name,
                         add_symbol=add_symbol,html_symbol=html_symbol,
                         latex_symbol=latex_symbol,ascii_symbol=ascii_symbol,
                         description=description)
    
    def tostring(self,fmt='unicode',norm=None,nsig=None,solidus=None,
                 mulsep=None,show_name=False,name=None,**kwds):
                     
        if name is None:
            name = Indexed.tostring(self,fmt)
        
        return super().tostring(fmt=fmt,norm=norm,nsig=nsig,solidus=solidus,
                                mulsep=mulsep,show_name=show_name,name=name)

def constant(name,toummy=None,splonk=None):
    """
    Finds an returns a constants from the constant library.
    
    Parameter
    ----------
        name: `str`, `Unit` or 1
            The name or alias of the constant.  
            
    Returns
    -------
    A `GummyConstant` or `JummyConstant` instance 
    
    ummy/immy Quantity instances can bre retrieved by setting `toummy` to True
    or GummyConstant.toummy to True.  splonk the returned value by setting
    `splonk` to True or GummyConstant.splonk to True
    """
    if isinstance(name,GummyConstant) or isinstance(name,JummyConstant):
        return name
    ret = JummyConstant._lib.get(name)
    if ret is None:
        ret = GummyConstant.get(name,exception=False)
    if ret is None:
        ret = JummyConstant.get(name,exception=False)
    if ret is None:
        raise ConstantNotFoundError('constant "' + str(name) + '" was not found')
        
    if splonk is None:
        splonk = GummyConstant.return_splonk
    if splonk:
        ret = ret.splonk()
    else:
        if toummy is None:
            toummy = GummyConstant.return_toummy
        if toummy:
            ret = ret.toummy()
        
    return ret

class search_constants_result(PrettyPrinter):
    """
    A `search_constants_result` instance emulates a list of constants returned 
    from a 'search_constants` function call, and pretty-prints the results to 
    the output
    """
    
    def __init__(self,constants):
        self.constants = constants
        
    def tostring(self,fmt='unicode',**kwds):
        if fmt == 'latex':
            txt = "<ul style=\"font-family: 'Times New Roman', Times, serif;font-size:1.2em\">\n"
        elif fmt == 'html':
            txt = "<ul>\n"
        else:
            txt = ''
        for u in self.constants:
            if fmt in ['latex','html']:
                txt += "<li>"
            try:
                txt += u.name + ' '
                
                if fmt == 'latex':
                    txt += _latex_math(u.tostring(fmt=fmt,show_name=True))
                else:
                    txt += u.tostring(fmt=fmt,show_name=True)
                    
                aliases = u.aliases
                if len(aliases) == 1:
                    txt += ', alias: ' + _mrtxt(aliases.pop(),fmt)
                if len(aliases) > 1:
                    txt += ', aliases: '
                    if u.short_name in aliases:
                        txt += _mrtxt(u.short_name,fmt) + ', '
                        aliases.remove(u.short_name)
                    txt += _mrtxt(', '.join(sorted(aliases,key=str.lower)),fmt)
                    
                saliases = u.shadowed_aliases
                if len(saliases) > 0:
                    if len(aliases) > 0:
                        txt += '; '
                    else:
                        txt += ', '
                    if len(saliases) == 1:
                        txt += 'shadowed alias: ' + _mrtxt(saliases.pop(),fmt)
                    else:
                        txt += 'shadowed aliases: ' + _mrtxt(', '.join(sorted(saliases,key=str.lower)),fmt)
            except:
                raise
                txt += '??'
                    
            if fmt == 'html' or fmt == 'latex':
                txt += '</li>\n'
            else:
                txt += '\n'
                
        txt = txt[:-1]
        if fmt in ['latex','html']:
            txt += '</ul>'
        
        return txt
    
    def __len__(self):
        return len(self.constants)
    
    def __getitem__(self,i):
        return self.constants[i]
    
    def __iter__(self):
        return iter(self.constants)
    
    def __reversed__(self):
        return reversed(self.constants)
    
    def __contains__(self,item):
        return item in self.constants

def search_constants(search=None,fmt=None,constants=None):
    """
    Prints a list of all loaded constant or all constants that match the search 
    terms.
    
    Parameters
    ----------
    search: `str` or `None`, optional
        A space separated list of search terms to case insentively match.
        If this is omitted or set equal to `None` then a list of all loaded
        constants will be printed.  The default is `None`.

    fmt: {'html','latex','unicode','ascii',`None`},optional
        The output format.  If `None`, then the `gummy.printer` value is used.
        If latex output is selected, Markdown is actually used with the unit
        symbols and conversion displayed using inline LaTeX.

    constants: `list` of `str`,optional
        A list of constants to print.  If this parameter is specified the values
        of the search and `show_all` parameters are ignored.
        
    Returns
    -------
    A `search_constants_result` instance which emulates a list of the returned
    constants and pretty-prints the results to the output or `None` if no
    constants are found.
    """
    
    if fmt is not None:
        fmt = fmt.lower().strip()
        if fmt == 'utf-8':
            fmt = 'unicode'
    
    if constants is None:
        while len(GummyConstant._builtins_to_import) > 0:
            import_module(GummyConstant._builtins_to_import.pop(),
                          GummyConstant.__module__)
            
        iconstants = {id(c):c for c in GummyConstant._builtin_lib.values()}
        iconstants.update({id(c):c for c in GummyConstant._lib.values()})
        iconstants.update({id(c):c for c in JummyConstant._builtin_lib.values()})
        iconstants.update({id(c):c for c in JummyConstant._lib.values()})
        iconstants = iconstants.values()
        
        if search is None:
            constants = iconstants
            if len(constants) == 0:
                print('no constants are loaded')
                return None
        else:
            constants = []
            for u in iconstants:
                s = set()
                for a in u.aliases:
                    s = s.union(set(a.lower().split()))
                for a in u.shadowed_aliases:
                    s = s.union(set(a.lower().split()))
                s = s.union(u.name.lower().split())
                s = s.union(set(u.symbol.lower().split()))
                s = s.union(set(u.symbol.lower().split()))
                if u.ascii_symbol is not None:
                    s = s.union(set(u.ascii_symbol.lower().split()))
                if u.description is not None:
                    s = s.union(set(u.description.lower().split()))
                        
                s = {i.strip(',.;:()') for i in s}
                
                srch = search.lower().split()
                ad = True
                
                for a in srch:
                    if a.strip(',.;') not in s:
                        ad = False
                        break
                if ad:
                    constants.append(u)
                    
            if len(constants) == 0:
                print('no constants found matching "' + search + '"')
                return None
            
        uf = []
        for u in constants:
            uf.append((u.name,u))
        constants = uf
    else:
        if (isinstance(constants,str) or isinstance(constants,GummyConstant) 
            or isinstance(constants,JummyConstant)):
            constants  = [constants]
        constants = ([constant(u) for u in constants])
                    
        constants = [(u.name,u) for u in constants]

    constants = sorted(constants,key=lambda u:u[0].lower())
        
    return search_constants_result([u[1] for u in constants])
        
        
def shadowed_constants(fmt=None,prnt=True):
    """
    Lists any constants which have a shadowed name or alias.  Constants may be 
    shadowed if the user has defined a new constant with the same name or alias
    as an existing constant.
    
    Parameters
    ---------
    fmt: {'html','latex','unicode','ascii',`None`},optional
        The output format.  If `None`, then the `gummy.printer` value is used.
        If latex output is selected, Markdown is actually used with the unit
        symbols and conversion displayed using inline LaTeX.
    """
    constants = {id(c):c for c in GummyConstant._builtin_lib.values()}
    constants.update({id(c):c for c in GummyConstant._lib.values()})
    constants.update({id(c):c for c in JummyConstant._builtin_lib.values()})
    constants.update({id(c):c for c in JummyConstant._lib.values()})
    constants = constants.values()
    constants = [u for u in constants if len(u.shadowed_aliases) > 0]
    return search_constants(fmt=fmt,constants=constants,prnt=prnt)



# -*- coding: utf-8 -*-

# module unit

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
Classes that for handling units are defined here:  Conversion, Unit,
_CompositeUnit, one, Quantity and QuantityArray
"""

import numpy as np
import weakref
import warnings
from .exceptions import (UnitLibError,CircularUnitConversionError,
                         UnitNotFoundError,IncompatibleUnitsError,
                         UnitLibNotFoundError,UnitWarning)
from .printing import PrettyPrinter,MetaPrettyPrinter
from .indexed import Indexed
from .unitparser import _UnitParser
from numbers import Rational,Integral,Number
from abc import ABCMeta
from fractions import Fraction
from decimal import Decimal

try:
    from mpmath import mp,mpf,rational
except:
    mp = mpf = rational = None

def unit(name,exception=True):
    """
    Finds an returns a `Unit` from the unit library.  This function is an 
    alias for the `Unit.unit` static method.
    
    Parameters
    ----------
        txt: `str`, `Unit` or 1
            This may be a string representing the unit.  The string can
            contain the name, short name or (if the unit was created
            with `add_symbol` set to `True`) the symbol of the unit or a
            combination of names and/or symbols of several different
            units.  Spaces or the character '*' represent multiplication,
            the character '/' represents division and the string '**'
            represents the power operator. For example txt can be:
                
            'kg m**2/s'
            
            or equivalently:
                
            'kilogram*metre*metre*second**-1' or '(kg/s)*m**2'.
            
            If a unit name contains a space, '*' or '/' character then the
            name must be enclosed in square brackets, e.g:
                
            [light year]
            
            If txt is a Unit instance that instance is returned.
            
        exception: `bool`, optional
            If this is `True` then a `UnitNotFoundError` or UnitLibError
            is raised if a unit is not found that matches `txt`.  If it is
            `False` and a unit is not found, then `Unit.unit` returns
            `None` without raising an exception.  The default is `True`.
            
    Returns
    -------
    A `Unit` instance or possibly None if the `exception` parameter is
    set to `True`.
    """

    return Unit.unit(name,exception=exception)

class Conversion:
    """
    Represents a unit conversion.  This class should only be used as arguments
    to `Unit` object initializers.  Each conversion should be associated with
    one and only one parent Unit; do not re-use conversion instances with more
    than one `Unit` instance.  
    
    The base `Conversion` class is defined by a single conversion factor.  
    More comlicated conversions overriding the `to` and `frm` methods should 
    inherit from the `NonlinearConversion` subclass and be used with units 
    that inherit from the `NonlinearUnit` subclass.
    
    Parameters
    ----------
    unit:  `str` or `Unit`
        the Unit that the parent Unit will be converted to.
        
    factor:  `float`, optional
        The conversion factor between the parent Unit and the new Unit:
        [value with new Unit] = factor * [value with parent Unit].  The
        conversion factor may not have a unit, but may be an ummy of gummy
        with not unit one.  The default value is 1.
    """
    linear = True
    
    def __init__(self,unit,factor=1):
        self._unit = unit
        if isinstance(factor,Fraction):
            factor = MFraction(factor)
        self.factor = factor
        
    def to(self,g):
        return g*self.factor
    
    def frm(self,g):
        return g/self.factor
        
    def chain(self,c):
        if c.linear:
            ret = Conversion(c.unit,self.factor*c.factor)
            ret.parent = self.parent
            return ret
        else:
            return c.rchain(self)
        
    @property
    def unit(self):
        if not isinstance(self._unit,Unit):
            self._unit = Unit.unit(self._unit)
        return self._unit
        
    def copy(self):
        r = Conversion(self.unit,self.factor)
        r.parent = self.parent
        return r
        
    def pow(self,e):
        r = Conversion(self.unit**e,self.factor**e)
        r.parent = self.parent**e
        return r


def _f_bop(f,rf,x,y):
    if isinstance(x,Quantity) and x.unit:
        if isinstance(y,Quantity):
            yunit = y.unit
            y = y.value
        else:
            yunit = one
        return f(x.unit,x.value,yunit,y,x.autoconvert)
    else:
        if isinstance(x,Quantity):
            xunit = x.unit
            x = x.value
        else:
            xunit = one
        return rf(y.unit,y.value,xunit,x,y.autoconvert)
    
def _f_ratio(f,x,y):
    if not isinstance(y,Quantity):
        x = x.convert(one).value
    elif not isinstance(x,Quantity):
        y = y.convert(one).value
    else:
        y = y.convert(x.unit).value
        x = x.value
    return (f(x,y),one)

class Unit(PrettyPrinter,Indexed):
    """
    Creating an instance of this class creates a representation of a physical
    unit and adds it to the unit library.  Once created the unit intance may
    be retrived by passing a string with the unit name or alias to the `unit`
    or `Unit.unit` functions.  Units can be multiplied and divided by other 
    Units or Quantities and raised to numerical powers.  Multiplying or 
    dividing a numerical value by a Unit will create a Quantity instance.
    
    Parameters
    ----------
        
    name:  `str`
        The name of the unit.  The name can be used access the unit with the
        `unit` function, but note that if you define a Unit with an
        identical name to a previously defined unit then the older name will
        be shadowed.
    
    symbol:  `str`
        A unicode symbol used when displaying the unit.  If the `add_symbol`
        parameter is set to `True`, then this symbol can also be used
        to access the unit with the `unit` function.
    
    conversion: `Conversion` or `None`, optional
        A conversion to another unit.  When creating units be careful to avoid
        circular conversions, i.e. you can define:
            
        Unit('inch','in',conversion=None)
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('ft',3))
        
        but not:
            
        Unit('inch','in',conversion=Conversion('yd',1/36))
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('ft',3))
        
        Note that an equivent and allowable way of defining the first set of 
        units above is:
            
        Unit('inch','in',conversion=None)
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('in',36))
        
        Either way will allow the free conversion between inches, feet, and
        yards.  You can also define the inch as:
            
        Unit('inch','in',conversion=('cm',2.54)
        
        (The inch is actually defined this way in the builtin unit library.)
    
    short_name: `str` or `None`
        a short name which can be used as an additional alias for the unit in
        the unit library
    
    add_symbol: `bool`, optional
        If this is `True`, then the symbol can be used to look up the unit
        in the unit library.  The default is `False`
        
    html_symbol, latex_symbol, ascii_symbol:  `str` or `Mone`, optional
        html, latex, and ascii versions of the symbol if they are different
        from the unicode representation of the symbol.
        
    description:  `str` or `None`, optional
        a description of the unit
    
    order:  `int` , optional
        When displaying composite derived units, the symbols with a lower
        `order` value will be displayed.  The default if -1.
        
        
    See Also
    --------
    The following are sub-classes of `Unit`:

    PrefixedUnit:  Creates a set of units with SI prefixes (..., kilo, mega, giga, ...)
    
    BinaryPrefixedUnit:  Creates a set of unit with binary prefixes (..., kibi, mebi, gibi, ...)
    
    LogUnit:  Logrithmic units (e.g. decibel or neper)
    
    OffsetUnit:  Units with an offset origin (degree Celsius or degree Fahrenheit)
    """
    # to speed loading of the module, only wait to load units until they are
    # needed.  If a unit cannot be found the modules below are loaded in 
    # reverse order.
    _builtins_to_import = ['..usunits','..miscunits','..siunits','..relunits']
    _builtin_lib = {}
    _lib = {}
    _open_lib = _lib
    
    _used_units = {}
    
    _ufunc_dict = {np.square: lambda x: (np.square(x.value),x.unit**2),
                   np.sqrt: lambda x: (np.sqrt(x.value),x.unit**0.5),
                   np.cbrt: lambda x: (np.cbrt(x.value),x.unit**(1/3)),
                   np.reciprocal: lambda x: (np.reciprocal(x.value),x.unit**-1),
                   np.absolute: lambda x: (np.absolute(x.value),x.unit),
                   np.negative: lambda x: (np.negative(x.value),x.unit),
                   np.positive: lambda x: (np.positive(x.value),x.unit),
                   np.conjugate: lambda x: (np.conjugate(x.value),x.unit),
                   np.rint: lambda x: (np.rint(x.value),x.unit),
                   np.floor: lambda x: (np.floor(x.value),x.unit),
                   np.ceil: lambda x: (np.ceil(x.value),x.unit),
                   np.trunc: lambda x: (np.trunc(x.value),x.unit),
                   np.add: lambda x,y: _f_bop(Unit._add,Unit._radd,x,y),
                   np.subtract: lambda x,y: _f_bop(Unit._sub,Unit._rsub,x,y),
                   np.multiply: lambda x,y: _f_bop(Unit._mul,Unit._rmul,x,y),
                   np.divide: lambda x,y: _f_bop(Unit._truediv,Unit._rtruediv,x,y),
                   np.floor_divide: lambda x,y: _f_bop(Unit._floordiv,Unit._rfloordiv,x,y),
                   np.mod: lambda x,y: _f_bop(Unit._mod,Unit._rmod,x,y),
                   np.power: lambda x,y: _f_bop(Unit._pow,Unit._rpow,x,y),
                   np.arctan2: lambda x,y: _f_ratio(np.arctan2,x,y)}
        
    solidus = True
    mulsep = False
    
    @staticmethod
    def unit(txt,exception=True):
        """
        Finds an returns a `Unit` from the unit library.
        
        Parameters
        ----------
            txt: `str`, `Unit` or 1
                This may be a string representing the unit.  The string can
                contain the name, short name or (if the unit was created
                with `add_symbol` set to `True`) the symbol of the unit or a
                combination of names and/or symbols of several different
                units.  Spaces or the character '*' represent multiplication,
                the character '/' represents division and the string '**'
                represents the power operator. For example txt can be:
                    
                'kg m**2/s'
                
                or equivalently:
                    
                'kilogram*metre*metre*second**-1' or '(kg/s)*m**2'.
                
                If a unit name contains a space, '*' or '/' character then the
                name must be enclosed in square brackets, e.g:
                    
                [light year]
                
                If txt is a Unit instance that instance is returned.
                
            exception: `bool`, optional
                If this is `True` then a `UnitNotFoundError` or UnitLibError
                is raised if a unit is not found that matches `txt`.  If it is
                `False` and a unit is not found, then `Unit.unit` returns
                `None` without raising an exception.  The default is `True`.
                
        Returns
        -------
        A `Unit` instance or possibly None if the `exception` parameter is
        set to `True`.
        """
        try:
            if isinstance(txt,Unit):
                return txt
            if txt is None or txt == 1:
                return one
            try:
                txt = txt.strip()
            except AttributeError:
                raise UnitNotFoundError('a unit may only be represented by a Unit object, the integer "1", or a string')
            if txt == '' or txt == '1':
                return one
            uu = Unit._used_units.get(txt)
            if uu is not None:
                return uu 
            ul = _UnitParser(txt).parse()
            if len(ul) == 1 and ul[0][1] == 1:
                if ul[0][0] == '1' or ul[0][0] == '' or ul[0][0] == 1:
                    return one
                else:
                    return Unit.get(ul[0][0])
            ull = []
            for u in ul:
                if u[0] == '1' or u[0] == '' or u[0] == 1:
                    un = one
                else:
                    un = Unit.get(u[0])
                if hasattr(un,'_getme'):
                    un = un._getme(ul,u[1])
                if isinstance(un,_CompositeUnit):
                    ull += [(uu[0],uu[1]*u[1]) for uu in un._units]
                else:
                    ull.append((un,u[1]))
            ret = _CompositeUnit(ull)
            Unit._used_units[txt] = ret
            return ret
        except UnitNotFoundError or UnitLibError:
            if (not txt.startswith('[') and not txt.endswith(']')):
                return Unit.unit('[' + txt + ']')
            elif exception:
                raise
            return None
        
    @staticmethod
    def reorder(txt):
        """
        This changes the order in which the symbols of composite derived units
        are displayed.

        Examples
        --------
        
        >>> print(Unit.unit('ft lb'))
        ft lb
        >>> print(Unit.unit('lb ft'))  #This is the same unit as above and displays identically
        ft lb
        >>> Unit.reorder('lb ft')  #Now the order will be changed when the unit is displayed
        >>> print(Unit.unit('ft lb'))
        lb ft
            
        """
        unit = Unit.unit(txt)
        if not isinstance(unit,_CompositeUnit):
            return
        
        ul = _UnitParser(txt).parse()
        for u in ul:
            if u[0] == '1' or u[0] == '' or u[0] == 1:
                un = one
            else:
                un = Unit.get(u[0])
            if hasattr(un,'_getme'):
                u[0] = un._getme(ul,u[1])
            else:
                u[0] = un
            if isinstance(un,_CompositeUnit):
                raise ValueError('aliases for composite units may not be used in Unit.reorder')
        unit._make_symbols(ul)
        
    @classmethod
    def _raise_not_found(cls,name):
        raise UnitNotFoundError('unit "' + str(name) + '" was not found')
        
    @staticmethod
    def format_latex(text):
        if text == '':
            return ''
        text = text.replace(' ','\\,')
        if text.startswith('\t\t'):
            return text[2:]
        if text.startswith('\\'):
            return text
        if text.startswith('\t'):
            return '\t\\mathrm{' + text[1:] + '}'
        return '\\mathrm{' + text + '}'
              
    _gconv = None
    _conversion = None
    _linear = True
    
    def __init__(self,name,symbol,conversion=None,short_name = None,
                 add_symbol=False,html_symbol=None,latex_symbol=None,
                 ascii_symbol=None,description=None,order = -1,**kwds):
        if conversion is not None and not isinstance(conversion,Conversion):
            raise ValueError('conversion is not an instance of the Conversion class')
        
        if symbol.startswith('\t\t'):
            raise ValueError('the symbol may not start with more than one tab character; only the latex_symbol may start with double tabs')
        
        super().__init__(name,symbol,short_name,add_symbol,html_symbol,
                         latex_symbol,ascii_symbol,description)
        
        self.conversion = conversion
        self.order = order
        
        self.parent = kwds.get('parent')
        
        #reset _used_units incase we are shadowing any unit definitions
        Unit._used_units = {} 

    @property
    def conversion(self):
        """
        Gets or sets the `Conversion` instance for the unit.
        
        ***This property is not intended to be used directly and setting this 
        property may cause problems***
        """
        return self._conversion
    @conversion.setter
    def conversion(self,c):
        self._conversion = c
        if c is not None:
            c.parent = self
            
    def _convs(self,h):
        if self in h:
            raise CircularUnitConversionError('a circular conversion chain was found nwith unit ' + self.name)
        h.add(self)
        
        try:
            c = self._conversion
            u = c.unit
            if isinstance(u,_CompositeUnit):
                una = u._units
                unb = {}
                for k,v in una:
                    if k in unb:
                        unb[k] += v
                    else:
                        unb[k] = v
                unb = {k:v for k,v in unb.items() if v != 0 and k is not one}
                if len(unb) == 0:
                    u = one
                else:
                    u = _CompositeUnit([(k,v) for k,v in unb.items()])
                    u._find_conv()
                c = c.chain(Conversion(u,1))
                     
        except UnitNotFoundError:
            warnings.warn('while searching conversions:  unit ' + self.name + ' should have a conversion to ' + self._conversion.unit_name + ' in library ' + str(self._conversion.unit_library) + ' but the second unit was not found',UnitWarning,stacklevel=3)
            return
        except UnitLibNotFoundError:
            warnings.warn('while searching conversions:  unit ' + self.name + ' should have a conversion to ' + self._conversion.unit_name + ' in library ' + str(self._conversion.unit_library) + ' but the library containing the second unit has not been loaded',UnitWarning,stacklevel=3)
            return
        
        if u._conversion is None:
            self._gconv = c
        else:
            r = u._convs(h)
            self._gconv = c.chain(r)
            
        return self._gconv
        
    @property
    def linear(self):
        """
        Gets a `bool` value indicating whether the Unit is linear.  If the
        unit is linear then any associated values will follow the standard 
        rules of arithmatic for Quantaties and the unit's Conversion is 
        defined by multiplying or dividing by a single conversion factor.
        Nonlinear units may have a more complicated conversion and may
        override the unusal operator methods.
        """

        return self._linear
                
    def _convtree(self):
        if self.conversion is not None and self._gconv is None:
            self._convs(set())
            
    def _convert(self,g,un):
        if self.conversion is not None and self.conversion.unit is un:
            return self.conversion.to(g)
            
        if self._gconv is not None and self._gconv._unit is un:
            return self._gconv.to(g)
            
        if un.conversion is not None and un.conversion.unit is self:
            return un.conversion.frm(g)
                
        if un._gconv is not None and un._gconv._unit is self:
            return un._gconv.frm(g)
        
        if self._gconv is not None and un._gconv is not None and un._gconv._unit is self._gconv._unit:
            g = self._gconv.to(g)
            g = un._gconv.frm(g)
            return g
            
        return None
        
    def convert(self,g,unit):
        """
        Converts a number from a quantity with the units `self` to `unit`.
        """
        unit = Unit.unit(unit)
        if unit is self:
            return g
            
        ret = self._convert(g,unit)
        if ret is not None:
            return ret
        self._convtree()
        unit._convtree()
        ret = self._convert(g,unit)

        unm = unit.short_name
        if ret is None:
            raise  IncompatibleUnitsError('no conversion found from unit ' + self.short_name + ' to unit ' + unm)
        return ret
    
    @property
    def is_dimensionless(self):
        """
        `bool`, read-only
        
        Returns `True` if a conversion exists between `self` and `one`, and
        `False` if not.
        """
        if self.conversion is not None and self.conversion.unit is one:
            return True
        if self._gconv is not None and self._gconv._unit is one:
            return True
        return False
        
    @property
    def units(self):
        """read-only
        
        Returns a list of the constituent units and their exponents, e.g. for
        kg m**2/s units would be [(kg, 1), (m, 2), (s, -1)].
        """
        return [(self,1)]
    
    @property
    def base(self):
        """read-only
        
        Returns the base unit.  That is we follow the chain of conversions
        starting with this units `Conversion` instance until we get to a unit
        that has no `Conversion` instance.  If this unit has no `Conversion`
        instance `self` is returned.
        """
        if self.conversion is None:
            return self
        self._convtree()
        return self._gconv._unit
    
    @staticmethod
    def _cancel(a,b=[],asign=1,bsign=1):
        # Takes .units lists a**asign and b**bsign and attempts to convert 
        # each unit in a into another unit either in a or b, and if successful 
        # cancels that unit.  Returns a tuple containing the  new .units 
        # list for the reduced composite unit a**asign * b**bsign as well as
        # the conversion factor between the unreduced and reduced composite unit.
        # In addition if both a an b contain dimensionless units then all
        # dimensionless units are converted to one.
        c = 1
        b = {i: bsign*j for i,j in b}
        for k,v in a:
            v *= asign
            if k in b:
                b[k] += v
            else:
                converted = False
                if not k.is_dimensionless:
                    for f in b.keys():
                        try:
                            c *= k.convert(1,f)**v
                            b[f] += v
                            converted = True
                            break
                        except  IncompatibleUnitsError:
                            pass
                if not converted:
                    b[k] = v
        un = []
        for k,v in b.items():
            if v != 0 and k is not one:
                uun,cc = k._cpow(v)
                un.append(uun[0])
                c *= cc
        return (un,c)
    
    def _add(self,s,vunit,v,aconv):
        if self is vunit:
            return (s + v,self)
        
        if not vunit.linear:
            return vunit._radd(v,self,s,not aconv)
        
        if aconv:
            s = self.convert(s,vunit)
            return (s + v,vunit)
        else:
            v = vunit.convert(v,self)
            return (s + v,self)
        
    def _radd(self,s,vunit,v,aconv):
        if self is vunit:
            return (v + s,self)
        
        if not vunit.linear:
            return vunit._add(v,self,s,not aconv)
        
        if aconv:
            s = self.convert(s,vunit)
            return (v + s,vunit)
        else:
            v = vunit.convert(v,self)
            return (v + s,self)
        
    def _sub(self,s,vunit,v,aconv):
        if self is vunit:
            return (s - v,self)
        
        if not vunit.linear:
            return vunit._rsub(v,self,s,not aconv)
        
        if aconv:
            s = self.convert(s,vunit)
            return (s - v,vunit)
        else:
            v = vunit.convert(v,self)
            return (s - v,self)
        
    def _rsub(self,s,vunit,v,aconv):
        if self is vunit:
            return (v - s,self)
        
        if not vunit.linear:
            return vunit._sub(v,self,s,not aconv)
        
        if aconv:
            s = self.convert(s,vunit)
            return (v - s,vunit)
        else:
            v = vunit.convert(v,self)
            return (v - s,self)
        
    def _mul_cancel(self,b):
        # self*b canceling units when possible
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,c[0])
        return (_CompositeUnit(d[0]),c[1]*d[1])
    
    def _mul(self,a,bunit,b,aconv):        
        if isinstance(b,Unit):
            bunit = b
            b = 1
        
        if not bunit.linear:
            return bunit._rmul(b,self,a,not aconv)

        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return ((a*b)*c,un)
    
    def _rmul(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if not bunit.linear:
            return bunit._mul(b,self,a,not aconv)

        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return ((b*a)*c,un)
        
    def _div_cancel(self,b):
        # self/b canceling units when possible        
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,b=c[0],asign=1,bsign=-1)
        return (_CompositeUnit(d[0]),c[1]*d[1])
        
    def _rdiv_cancel(self,b):
        # b/self canceling units when possible
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,b=c[0],asign=-1,bsign=1)
        return (_CompositeUnit(d[0]),c[1]*d[1])
        
    def _truediv(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if not bunit.linear:
            return bunit._rtruediv(b,self,a,not aconv)

        if aconv:
            un,c = self._div_cancel(bunit)
        else:
            un,c = bunit._rdiv_cancel(self)
        
        return ((a/b)*c,un)
    
    def _rtruediv(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if aconv:
            un,c = self._rdiv_cancel(bunit)
        else:
            un,c = bunit._div_cancel(self)
        
        return ((b/a)*c,un)
    
            
    def _floordiv(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if not bunit.linear:
            return bunit._rtruediv(b,self,a,not aconv)

        if aconv:
            un,c = self._div_cancel(bunit)
        else:
            un,c = bunit._rdiv_cancel(self)
        
        return ((a // b)*c,un)
    
    def _rfloordiv(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if aconv:
            un,c = self._rdiv_cancel(bunit)
        else:
            un,c = bunit._div_cancel(self)
        
        return ((b // a)*c,un)
        
    def _mod(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if self is bunit:
            return (a % b,self)
        if not bunit.linear:
            return bunit._rmod(b,self,a,not aconv)
        try:
            if aconv:
                return (self.convert(a,bunit) % b,bunit)
            return (a % bunit.convert(b,self),self)
        except IncompatibleUnitsError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())

    def _rmod(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if self is bunit:
            return (b % a,self)
        try:
            if aconv:
                return (b % self.convert(a,bunit),bunit)
            return (bunit.convert(b,self) % a,self)
        except IncompatibleUnitsError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())

    def _cpow(self,v):
        return ([[self,v]],1)
    
    def _pow(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if not bunit.linear:
            return bunit._rpow(b,self,a,not aconv)
        
        if bunit is not one:
            try:
                b = bunit.convert(b,one)
            except IncompatibleUnitsError:
                raise IncompatibleUnitsError('the exponent is not dimensionless')
                
        un,c = self._cpow(b)
        un = _CompositeUnit(un)
        return ((a**b)*c,un)
    
    def _rpow(self,a,bunit,b,aconv):
        if isinstance(b,Unit):
            bunit = b
            b = 1
            
        if self is not one:
            try:
                a = self.convert(a,one)
            except IncompatibleUnitsError:
                raise IncompatibleUnitsError('the exponent is not dimensionless')
        un,c = bunit._cpow(a)
        un = _CompositeUnit(un)
        return ((b**a)*c,un)
    
    def _ufunc(self,func,*args,**kwds):
        if self.linear:
            try:
                nl = next(a for a in args 
                          if isinstance(a,Quantity) and not a.unit.linear)
                return nl.unit._ufunc(func,*args,**kwds)
            except StopIteration:
                pass
            
        try:
            f = type(self)._ufunc_dict[func]
        except KeyError:
            raise NotImplementedError()
        return f(*args,**kwds)
    
    def _neg(self,a):
        return (-a,self)
    
    def _pos(self,a):
        return (+a,self)
    
    def _abs(self,a):
        return (abs(a),self)
    
    def __mul__(self,v):
        if not isinstance(v,Unit):
            return Quantity._make(v,unit=self)
        
        if not v.linear:
            return v.__rmul__(self)
        
        if v is one:
            return self
        
        return _CompositeUnit(self.units + v.units)
    
    def __rmul__(self,v):
        if not isinstance(v,Unit):
            return Quantity._make(v,unit=self)
        
        if not v.linear:
            return v.__mul__(self)
        
        if v is one:
            return self
        
        return _CompositeUnit(self.units + v.units)
        
    def __truediv__(self,v):
        if v is one:
            return self
        
        if not isinstance(v,Unit):
            return Quantity._make(1/v,unit=self)
            
        if not v.linear:
            return v.__rtruediv__(self)
            
        vi = [(e[0],-e[1]) for e in v.units]
        return _CompositeUnit(self.units + vi)

    def __rtruediv__(self,v):
        if v is one or v == 1:
            return _CompositeUnit([(e[0],-e[1]) for e in self.units])
        
        return Quantity._make(v,unit=self**-1)
        
    def __pow__(self,v):
        if v == -1:
            return _CompositeUnit([(e[0],-e[1]) for e in self.units])
        
        if v == 0:
            return one
        
        if v == 1:
            return self
        
        if v != float(v):
            raise TypeError('unsupported operand type(s) for **: Unit and ' + str(v))

        return _CompositeUnit([(e[0],v*e[1]) for e in self.units])
        
    
class _CompositeUnit(Unit):
    # Instances of this class represent composite derived unit.
    
    living_units = weakref.WeakValueDictionary()
    
    parent = None
    
    @staticmethod
    def _unicode_super(e):
        if int(e) == e:
            se = str(int(e))
            so = ''
            for s in se:
                if s == '-':
                    so += '\u207B'
                if s == '0':
                    so += '\u2070'
                if s == '1':
                    so += '\u00B9'
                if s == '2':
                    so += '\u00B2'
                if s == '3':
                    so += '\u00B3'
                if s == '4':
                    so += '\u2074'
                if s == '5':
                    so += '\u2075'
                if s == '6':
                    so += '\u2076'
                if s == '7':
                    so += '\u2077'
                if s == '8':
                    so += '\u2078'
                if s == '9':
                    so += '\u2079'
            return so
        return '**' + _CompositeUnit._format_sscript(e,parenth=True)
    
    @staticmethod
    def _format_sscript(e,parenth=False):
        if isinstance(e,Rational) and not isinstance(e,Integral):
            ret = str(e.numerator) + '/' + str(e.denominator)
            if parenth:
                ret = '(' + ret + ')'
            return ret
        mf = np.modf(e)
        if mf[0] == 0:
            return '{:.0f}'.format(e)
        if mf[0] == 0.5:
            return '{:.0f}'.format(2*e) + '/2'
        if mf[0] == 0.25:
            return '{:.0f}'.format(4*e) + '/4'
        return str(e).rstrip('0.')
        
    @staticmethod
    def _iaddsym(txt,s,e,fmt,f):
        if txt.startswith('\t'):
            txt = txt[1:]
        if e == 0:
            return txt
        if f == 1 and e < 0:
            return txt
        if f == 2:
            if e > 0:
                return txt
            e = -e
            
        if fmt == 'html':
            if txt != '':
                txt += '&middot;'
            txt += s
            if e != 1:
                txt += '<sup>' + _CompositeUnit._format_sscript(e) + '</sup>'
        elif fmt == 'latex':
            if txt != '':
                txt += r'\cdot'
            txt += s
            if e != 1:
                txt += '^{' + _CompositeUnit._format_sscript(e) + '}'
        elif fmt == 'ascii':
            if txt != '':
                txt += '*'
            txt += s
            if e != 1:
                txt += '**' + _CompositeUnit._format_sscript(e,parenth=True)
        else:
            if txt != '':
                txt += ' '
            txt += s
            if e != 1:
                txt += _CompositeUnit._unicode_super(e)
            
        return txt
        
    def _addsym(txt,u,fmt=None,f=0):
        return _CompositeUnit._iaddsym(txt,u[0].tostring(fmt=fmt),u[1],fmt,f)
    
    def __new__(cls,ul):
        
        # ul is a list of tupels in each tuple a (non _CompositeUnit) Unit
        # instance followed by an exponent.
        ul = [u for u in ul if u[0] is not one and u[1] != 0]
        
        if len(ul) == 0:
            return one
        
        if len(ul) == 1 and ul[0][1] == 1:
            return ul[0][0]
        
        ul = sorted(ul,key = lambda x: (x[0].order,-x[1]))
        
        uid = tuple([(id(u[0]),u[1]) for u in ul])
        ret = _CompositeUnit.living_units.get(uid)
        if ret is not None:
            return ret
        
        linear = True
        for u,e in ul:
            if not u.linear:
                linear = False
                try:
                    cls = u.get_composite(ul)
                except NotImplementedError:
                    raise IncompatibleUnitsError('unit ' + u.tostring() + ' may not be combined with other units')
                break
        
        ret = object.__new__(cls)
        
        ret._linear = linear
        ret.order = 100
        ret._units = ul
        _CompositeUnit.living_units[uid] = ret
        ret._oconv = None
        ret._have_conversion = False
        ret._aliases = set()
        
        nm = ''
        snm = ''
        for u in ul:          
            nm += ' '
            snm += ' '
            nmi = u[0].name
            if ' ' in nmi and (not nmi[0].startswith('[') or not nmi.endswith(']')):
                nmi = '[' + nmi + ']'
            snmi = u[0].short_name
            if ' ' in snmi and (not snmi[0].startswith('[') or not snmi.endswith(']')):
                snmi = '[' + snmi + ']'
            if u[1] == 1:
                    nm += nmi
                    snm += snmi
            else:
                    nm += nmi + '**' + str(u[1])
                    snm += snmi + '**' + str(u[1])
        
        ret.name = nm.strip()
        ret.short_name = snm.strip()
        
        ret._make_symbols(ul)
        
        return ret
    
    def __init__(self,ul):
        pass
        
    def _make_symbols(self,ul):
        sym = ''
        hsym = ''
        lsym = ''
        asym = ''
        symsn = ''
        symsd = ''
        hsymsn = ''
        hsymsd = ''
        lsymsn = ''
        lsymsd = ''
        asymsn = ''
        asymsd = ''
        
        nd = 0
        for u in ul:
            sym = _CompositeUnit._addsym(sym,u)
            hsym = _CompositeUnit._addsym(hsym,u,fmt = 'html')
            lsym = _CompositeUnit._addsym(lsym,u,fmt = 'latex')
            asym = _CompositeUnit._addsym(asym,u,fmt = 'ascii')
            
            symsn = _CompositeUnit._addsym(symsn,u,f=1)
            hsymsn = _CompositeUnit._addsym(hsymsn,u,fmt = 'html',f=1)
            lsymsn = _CompositeUnit._addsym(lsymsn,u,fmt = 'latex',f=1)
            asymsn = _CompositeUnit._addsym(asymsn,u,fmt = 'ascii',f=1)
            
            symsd = _CompositeUnit._addsym(symsd,u,f=2)
            hsymsd = _CompositeUnit._addsym(hsymsd,u,fmt = 'html',f=2)
            lsymsd = _CompositeUnit._addsym(lsymsd,u,fmt = 'latex',f=2)
            asymsd = _CompositeUnit._addsym(asymsd,u,fmt = 'ascii',f=2)
            
            if u[1] < 0:
                nd += 1
            
        self.sym = sym
        self.hsym = hsym
        self.lsym = lsym
        self.asym = asym
        
        if symsn == '' and nd > 0:
            symsn = '1'
            hsymsn = '1'
            lsymsn = '1'
            asymsn = '1'
        
        if nd > 0:
            if nd > 1:
                symsd = '(' + symsd + ')'
                hsymsd = '(' + hsymsd + ')'
                lsymsd = '(' + lsymsd + ')'
                asymsd = '(' + asymsd + ')'
            syms = symsn + '/' + symsd
            hsyms = hsymsn + '/' + hsymsd
            lsyms = lsymsn + '/' + lsymsd
            asyms = asymsn + '/' + asymsd
        else:
            syms = symsn
            hsyms = hsymsn
            lsyms = lsymsn
            asyms = asymsn
        self.syms = syms
        self.hsyms = hsyms
        self.lsyms = lsyms
        self.asyms = asyms
        
    @property
    def units(self):
        """
        read-only

        Returns a list of the constituent units and their exponents, e.g. for
        kg m**2/s units would be [(kg, 1), (m, 2), (s, -1)].
        """
        return self._units
        
    def tostring(self,fmt=None,solidus=None,mulsep=None,**kwds):
        """
        Returns a string containing the symbol for the unit the format given by
        the keyword fmt which may be set to a string the values 'html', 'latex', 
        'ascii', or 'unicode'.
        
        If solidus is `True` a slash will be used instead of negative exponents
        and if `mulsep` is `True` a mid-dot or asterix will be used instead of a space
        to separate units.
        """
        if solidus is None:
            solidus = self.solidus
            
        if mulsep is None:
            mulsep = self.mulsep
            
        if fmt is None or fmt == 'utf-8' or fmt == 'unicode':
            if solidus:
                ret = self.syms
            else:
                ret = self.sym
            if not mulsep:
                ret = ret.replace('\u00B7',' ')
            return ret
        elif fmt == 'html':
            if solidus:
                ret = self.hsyms
            else:
                ret = self.hsym
            if not mulsep:
                ret = ret.replace('&middot;','&thinsp;')
        elif fmt == 'latex':
            if solidus:
                ret = self.lsyms
            else:
                ret = self.lsym
            if not mulsep:
                ret = ret.replace(r'\cdot',r'\,')
            return ret
        elif fmt == 'ascii':
            if solidus:
                ret = self.asyms
            else:
                ret = self.asym
            if not mulsep:
                ret = ret.replace('**','#@@#')
                ret = ret.replace('*',' ')
                ret = ret.replace('#@@#','**')
        else:
            raise ValueError('format ' + str(fmt) + ' is unrecognized')
        return ret
        
    def _find_conv(self):    
        cg = None
        cgun = {}
        for un,e in self._units:
            un._convtree()
                    
            if un._gconv is None:
                if un in cgun:
                    cgun[un] += e
                else:
                    cgun[un] = e
            else:
                p = un._gconv.pow(e)
                if cg is None:
                    cg = p
                else:
                    cg = cg.chain(p)
                for t in p.unit.units:
                    if t[0] in cgun:
                        cgun[t[0]] += t[1]
                    else:
                        cgun[t[0]] = t[1]
                        
        cgun = [(k,v) for k,v in cgun.items() if v != 0]
        
        cgun = _CompositeUnit(cgun)
        if cgun is not self:
            if cg is None:
                cg = Conversion(cgun,1)
            else:
                cg._unit = cgun
            cg.parent = self
            self._gconv = self._conversion = cg
        else:
            self._gconv = self._conversion = None

        self._have_conversion = True

    
    def _cpow(self,v):
        c = 1
        un = []
        for u,e in self.units:
            uun,cc = u._cpow(v*e)
            un.append(uun[0])
            c *= cc
        return (un,c)

    @property
    def conversion(self):
        """
        Gets or sets the conversion object for the unit.  
        
        ***This property is not intended to be used directly and setting this 
        property may cause problems***
        """
        if not self._have_conversion:
            self._find_conv()
        return self._conversion
    
    @property
    def aliases(self):
        """read-only
        
        Returns a set of the unshadowed aliases of this unit.  To add aliases
        to the unit use the `Unit.alias` static method.
        """
        return {a for a in self._aliases if (Unit.unit('[' + a + ']',exception=False) is self and a != self.name)}
    
    @property
    def shadowed_aliases(self):
        """
        Returns a set of the shadowed aliases of this unit.
        """
        sha = {a for a in self._aliases if Unit.unit('[' + a + ']',exception=False) is not self}
        return sha
    
class _One(Unit):
    """
    The only instance of this class should be the unit `one`.
    """
    def __mul__(self,v):
        if isinstance(v,Unit):            
            return v
        return Quantity._make(v)
        
    def __truediv__(self,v):
        if isinstance(v,Unit):            
            return v**-1
        return Quantity._make(1/v)

    def __rtruediv__(self,v):
        if isinstance(v,Unit):            
            return v
        return Quantity._make(v)
        
    def __pow__(self,v):
        return self
    
    def _cpow(self,v):
        return ([[self,1]],1)
    
    def _repr_html_(self):
        return 'one'
            
    def _repr_latex_(self):
        return 'one'
    
    def __str__(self):
        return 'one'
            
    def __repr__(self):
        return 'one'
    
    @property
    def is_dimensionless(self):
        return True
    
with Unit._builtin():
    one = _One('1','',add_symbol=False)

class MetaQuantity(MetaPrettyPrinter,ABCMeta):
    pass

class Quantity(PrettyPrinter,Number,metaclass=MetaQuantity):
    """
    Instances of this class represent a quantity with a value and a unit.
    The behavior of Quantity instances under mathematical operations with
    other Quanitity object or numerical values depends on the unit. E.g. 
    an interger of float may be added to `Quantity(1,unit='%')` but not to 
    `Quantity(1,unit='m/s')`.  For operations involving only linear units, the
    units will be automatically converted to facilitate the operation, e.g.
    `Quantity(1,unit='psi')` may be added to `Quantity(1,unit='psi')` but not
    `Quantity(1,unit='db(uPa)'.  Manual unit conversions can be realized by
    calling the `Quantity.convert` method, or in place by setting the 
    `Quantity.unit` property.
    
    Quantity instances may be created directly or by multiplying or dividing 
    a number by a `Unit` instance: `Quantity(2.2,unit='cm')` is equivalent to
    `2.2 * unit('cm')`.

    Parameters
    ----------

    value: numeric (including `ummy`)
        the value of the Quantity

    unit:  `str` or `Unit`
        The `Unit` instance representing the unit of the Quantity or a string
        that references the `Unit` instance.
    """
    
    splonk_func_ret = False
    
    _arraytype = None
    
    @classmethod
    def _make(cls,value,unit=one):
        if (isinstance(value,np.ndarray) or isinstance(value,list) or
            isinstance(value,tuple)) and cls._arraytype is not None:
            return cls._arraytype(value,unit=unit)
        return cls(value,unit=unit)
            
    
    @staticmethod
    def _add_unit_sp(fmt,unit):
        if unit is None or unit is one:
            return ''
            
        if unit == '':
            return ''
            
        if unit.startswith('\t'):
            unit = unit[1:]
        else:
            if fmt == 'latex':
                unit = r'\:' + unit
            elif fmt == 'html':
                unit = '&nbsp;' + unit
            else:
                unit = ' ' + unit
            
        return unit
    
    def __init__(self,value,unit=one):
        try:
            hash(value)
        except TypeError:
            raise TypeError('value is unhashable')
        self._value = value
        if isinstance(unit,Unit):
            self._unit = unit
        else:
            self._unit = Unit.unit(unit)
        self._old = None
        self.autoconvert = False
        
    @property
    def value(self):
        return self._value
           
    @property
    def unit(self):
        """
        Gets or sets the unit for the Quantity.
        
        If this property is set, a unit conversion will be performed.  The value 
        it is set to may be a string, `None`, a `Unit` object, or the integer 1.
        Both 1 and `None` will be interpreted as the Unit instance `one`. A
        `NoUnitConversionFoundError` will be raised if the unit conversion is
        not possible.
        
        Example
        -------
            
        >>> x = Quantity(0.001,unit='V')
        >>> x
        0.001 V
        >>> x.unit = 'uV'
        >>> x
        1000.0 uV
        """
        return self._unit
    @unit.setter
    def unit(self,unit):     
        #Always convert back to the original units so we don't accumulate
        #rounding errors.
        if self._old is None:
            self._old = (self.value,self.unit)
        else:
            self._value,self._unit = self._old
            
        unit = Unit.unit(unit)
        value = self._unit.convert(self.value,unit)
        
        self._value = value
        self._unit = unit
        return
        
    def convert(self,unit):
        """
        Returns a copy of the Quantity with converted units.  A 
        `NoUnitConversionFoundError` will be raised if the unit conversion is
        not possible.
        
        
        Parameters
        ----------
        unit:  `str` or `Unit`
            The unit for the `x` value and if `uunit` is `None`, the
            uncertainty.  It must be string, None, a `Unit` object, or the
            integer 1.  Both 1 and `None` will be interpreted as the Unit
            instance `one`.

        uunit `str`, `Unit` or None, optional
            The unit for the uncertainty `U`.  If this is `None` then `U`
            will have the same units as `x`.  The default is `None`.
        """
        unit = Unit.unit(unit)
        value = self.unit.convert(self.value,unit)
        return type(self)(value,unit=unit)
        
    def reduce_unit(self):
        """
        Cancels factors in a Quantity's unit when possible.  This modifies the
        calling gummy and returns `None`.
        
        Example
        -------
        
        >>> x = Quantity(5,unit='mm/m')
        >>> x.reduce_unit()
        >>> x
        0.005
        """
        un,f = self._unit._mul_cancel(one)
        self._unit = un
        self._value *= f
    
    @property
    def c(self):
        """
        This read-only property is used as a conversion flag during calculations.
        When an arithmetic operation is carried out between two Quantaties with
        different units, a unit conversion on one of the input quantities may be
        required to complete the calculation.  Attach this flag to the unit that
        you prefer be converted.

        Examples
        --------
        
        >>> a = Quantity(1,unit='cm')
        >>> b = Quantity(2,unit='mm')
        >>> a + b
        1.2 cm
        >>> a.c + b
        12 mm
        >>> a + b.c
        1.2 cm
        >>> a*b
        0.2 cm**2
        >>>a.c*b
        20 mm**2
        """
        c = Quantity(self.value,self.unit)
        c.autoconvert = True
        return c
 
    def tostring(self,fmt=None,**kwds):
        """
        tostring(fmt='unicode')
        
        returns a string representation of the Quantity.  fmt may be "unicode",
        "html","latex" or "ascii".  The default is "unicode".
        """
        unit = self._unit.tostring(fmt=fmt,**kwds)
        unit = self._add_unit_sp(fmt,unit)
        if isinstance(self.value,PrettyPrinter):
            value = self.value.tostring(fmt=fmt,**kwds)
        else:
            value = str(self.value)
        return value + unit
    
    def copy(self,tofloat=False):
        """
        copy(tofloat=False)
        
        returns a copy of self.  If tofloat is True, the self.value will be
        converted to float.  The default is False.
        """
        if tofloat:
            try:
                return type(self)(self.value.tofloat(),unit=self.unit)
            except:
                return type(self)(float(self._value),unit=self._unit)
            
        try:
            return type(self)(self.value.copy(),unit=self.unit)
        except:
            return type(self)(self.value,self.unit)
        
    def tofloat(self):
        """
        returns a copy of self with value float(self.value) equivalent to 
        copy(tofloat=True)
        """
        return self.copy(tofloat=True)
    
    def totuple(self):
        """
        returns the tuple (self.value,self.unit)
        """
        return (self.value,self.unit)
    
    def tobaseunit(self):
        """
        Returns a Quantity equal to self converted to unit self.unit.base
        """
        return self.convert(self.unit.base)
    
    def splonk(self):
        """
        returns self.value if self.unit is one else returns self
        """
        if self.unit is one:
            try:
                return self.value.splonk()
            except AttributeError:
                return self.value
        return self
        
    def _bop(self,v,f):
        if issubclass(type(v),type(self)):
            make =  v._make
        else:
            make =  self._make
            
        if isinstance(v,Quantity):
            vunit = v._unit
            v = v.value
            aconv = self.autoconvert
        else:
            vunit = one
            aconv = False
            
        r,runit = f(self.value,vunit,v,aconv)
            
        return make(r,unit=runit)
        
    def __add__(self, v):
        return self._bop(v,self.unit._add)
                
    def __radd__(self, v):
        return self._bop(v,self.unit._radd)
    
    def __sub__(self, v):
        return self._bop(v,self.unit._sub)
                
    def __rsub__(self, v):
        return self._bop(v,self.unit._rsub)
    
    def __mul__(self, v):
        return self._bop(v,self.unit._mul)
                
    def __rmul__(self, v):
        return self._bop(v,self.unit._rmul)
    
    def __truediv__(self, v):
        return self._bop(v,self.unit._truediv)
                
    def __rtruediv__(self, v):
        return self._bop(v,self.unit._rtruediv)
    
    def __floordiv__(self, v):
        return self._bop(v,self.unit._floordiv)
                
    def __rfloordiv__(self, v):
        return self._bop(v,self.unit._rfloordiv)
    
    def __pow__(self, v):
        return self._bop(v,self.unit._pow)
                
    def __rpow__(self, v):
        return self._bop(v,self.unit._rpow)
    
    def __mod__(self, v):
        return self._bop(v,self.unit._mod)
                
    def __rmod__(self, v):
        return self._bop(v,self.unit._rmod)                
        
    def __neg__(self):
        r,runit = self.unit._neg(self.value)
        return self._make(r,unit=runit)
        
    def __pos__(self):
        r,runit = self.unit._pos(self.value)
        return self._make(r,unit=runit)
        
    def __abs__(self):
        r,runit = self.unit._abs(self.value)
        return self._make(r,unit=runit)
    
    def _cmp(self,v,f):
        if isinstance(v,Quantity):
            if self._unit is not v.unit:
                try:
                    v = v.convert(self.unit)
                except IncompatibleUnitsError:
                    raise IncompatibleUnitsError('Quantities with incompatible units cannot be ordered')
            return f(self.value,v.value)
        else:
            try:
                s = self.convert(one)
                return f(s.value,v)
            except IncompatibleUnitsError:
                raise IncompatibleUnitsError('Quantities with incompatible units cannot be ordered')
        
    def __eq__(self, v):
        try:
            return self._cmp(v,lambda x,y: x == y)
        except IncompatibleUnitsError:
            return False
    
    def __ne__(self, v):
        try:
            return self._cmp(v,lambda x,y: x != y)
        except IncompatibleUnitsError:
            return True
        
    def __lt__(self, v):
        return self._cmp(v,lambda x,y: x < y)
    
    def __le__(self, v):
       return self._cmp(v,lambda x,y: x <= y)
        
    def __gt__(self, v):
        return self._cmp(v,lambda x,y: x > y)
        
    def __ge__(self, v):
        return self._cmp(v,lambda x,y: x >= y)
    
    def _ufunc(self,func,*args,**kwds):
        # handles numpy functions applied to Quantity arguments
        
        s = self
        for a in args:
            if issubclass(type(a),type(s)):
                s = a
        make = s._make
        
        try:
            x,xunit = self.unit._ufunc(func,*args,**kwds)
            return make(x,unit=xunit)
        except NotImplementedError:
            try:
                x = [a if not isinstance(a,Quantity) else a.convert(one).value 
                     for a in args]
                if self.splonk_func_ret:
                    return func(*x,**kwds)
                return make(func(*x,**kwds))
            except IncompatibleUnitsError:
                raise NotImplementedError()
      
    def __array_ufunc__(self,ufunc,method,*args,**kwds):
        # handles numpy ufunc's applied to Quantity arguments
        if method != '__call__':
            return None
        
        return self._ufunc(ufunc,*args,**kwds)
    
    def __array_function__(self,func,method,*args,**kwds):        
        return self._ufunc(func,*args,**kwds)
    
    def __float__(self):
        s = self.convert(1)
        return float(s.value)

    def __complex__(self):
        s = self.convert(1)
        return complex(s.value)
    
    def __bool__(self):
        return self != 0
    
    @property
    def real(self):
        if self.unit.linear:
            s = self
        else:
            s = self.tobaseunit()
         
        try:
            v = s.value.real
        except:
            v = float(s.value)
        return type(self)(v,unit=s.unit)
     
    @property
    def imag(self):
        if self.unit.linear:
            s = self
        else:
            s = self.tobaseunit()
        
        try:
            v = s.value.imag
        except:
            v = 0
        return type(self)(v,unit=s.unit)
    
    def __hash__(self):
        r = self.tobaseunit().totuple()
        if r[1] is one:
            return hash(r[0])
        return hash(r)
    
    def conjugate(self):
        try:
            v = self.value.conjugate()
        except:
            v = complex(float(self.value))
        return type(self)(v,unit=self.unit)


class QuantityArray(Quantity):
    """
    A subclass of Quantity.  Instance of this class represent a list, tuple, 
    or numpy array of values all with the same unit.  Elements of the array 
    are returned as `Quantity` instances.  Instances of this class can be
    created directly or by multiplying a list, tuple, or numpy array by a
    `Unit` instance.
    
    Parameters
    ----------

    value: `list`, `tuple` or `ndarray`
        the value of the Quantity

    unit:  `str` or `Unit`
        The `Unit` instance representing the unit of the Quantity or a string
        that references the `Unit` instance.
    """
    
    _elementtype = Quantity
    
    def __len__(self):
        return len(self.value)
    
    def __getitem__(self,key):
        return self._elementtype(self.value[key],unit=self.unit)

    def __contains__(self,quantity):
        if isinstance(quantity,self._elementtype):
            if self._unit is not quantity.unit:
                try:
                    quantity = quantity.convert(self.unit)
                    return quantity.value in self.value
                except:
                    return False
        else:
            try:
                s = self.convert(one)
                return quantity in s.value
            except:
                return False
            
    def __iter__(self):
        for e in self.value:
            yield self._elementtype(e,unit=self.unit)
            
    def __array__(self):
        return type(self)(np.asarray(self.value),unit=self.unit)
    
    
Quantity._arraytype = QuantityArray

        
class MFraction(Fraction):
    """
    A fraction.Fraction sub-class that works with Decimal and mpmath.mpf objects
    """
    
    @classmethod
    def fromnum(cls,x):
        return MFraction(Fraction(x))
    
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and not (isinstance(args[0],str) 
                                   or isinstance(args[0],Fraction)):
            return args[0]
        ret = super(MFraction, cls).__new__(cls, *args, **kwargs)
        if ret.denominator == 1:
            return ret.numerator
        return ret
    
    def _mpmath_(self,p,r):
        return rational.mpq(self.numerator,self.denominator)
    
    def todecimal(self):
        return Decimal(self.numerator)/Decimal(self.denominator)
    
    def __add__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__add__(v)
        return MFraction(super().__add__(v))
    
    def __radd__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__radd__(v)
        return MFraction(super().__radd__(v))
    
    def __sub__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__sub__(v)
        return MFraction(super().__sub__(v))
    
    def __rsub__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rsub__(v)
        return MFraction(super().__rsub__(v))
    
    def __mul__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__mul__(v)
        return MFraction(super().__mul__(v))
    
    def __rmul__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rmul__(v)
        return MFraction(super().__rmul__(v))
    
    def __truediv__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__truediv__(v)
        return MFraction(super().__truediv__(v))
    
    def __rtruediv__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rtruediv__(v)
        return MFraction(super().__rtruediv__(v))
        
    def __pow__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__pow__(v)
        return MFraction(super().__pow__(v))
    
    def __rpow__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rpow__(v)
        if isinstance(v,Fraction):
            return MFraction(v).__pow__(self)
        return MFraction(super().__rpow__(v))
        
    def __floordiv__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__floordiv__(v)
        return MFraction(super().__floordiv__(v))
        
    def __rfloordiv__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rfloordiv__(v)
        return MFraction(super().__rfloordiv__(v))
        
    def __mod__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__mod__(v)
        return MFraction(super().__mod__(v))
    
    def __rmod__(self,v):
        if isinstance(v,Decimal):
            return self.todecimal().__rmod__(v)
        return MFraction(super().__rmod__(v))
    
    def __abs__(self):
        return MFraction(super().__abs__())
    
    def __neg__(self):
        return MFraction(super().__neg__())
    
    def __pos__(self):
        return MFraction(super().__pos__())




# -*- coding: utf-8 -*-

# module test_ubreakdown

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

import metrolopy as uc
import numpy as np
import unittest

from metrolopy.tests.common import rand,make_gummy

class TestUbreakdown(unittest.TestCase):
    def test_ubreakdown(self,n=None,prnt=False,budget=False):
        """
        Test gummy.ubreakdown and the budget class.
        """
        uc.gummy.p_method = None
        
        if n is None:
            if prnt or budget:
                n = 100
            else:
                n = 1000
            
        for i in range(n):
            k = rand.randint(1,6)
            
            if rand.randint(2):
                unt = '1'
            else:
                unt = None
            g,x,u,dof,unit,utype = zip(*[make_gummy(unit=unt) for y in range(k)])
            
            for i,y in enumerate(g):
                if y.u != 0:
                    self.assertTrue(y.utype is utype[i])
            
            c = rand.rand()*0.9 + 0.1
            if rand.randint(2):
                c = -c
            
            def f(*p):
                r = c
                for y in p:
                    r *= y
                return r
            
            def d(k,*p):
                r = c
                for i,y in enumerate(p):
                    if i != k:
                        r *= y
                return r
            
            gr = f(*g)
            
            ut = None
            for q in g:
                if abs(gr.correlation(q)) == 1:
                    ut = q.utype
            if ut is not None:
                self.assertTrue(gr.utype is ut)
            
            x = list(x)
            u = list(u)
            lb = 'lb' in str(gr.unit)
            kg = 'kg' in str(gr.unit)
            for i,un in enumerate(unit):
                if 'lb' in un and not lb:
                    x[i] *= 0.45359237
                    u[i] *= 0.45359237
                if 'kg' in un and not kg and lb:
                    x[i] *= 0.45359237**4
                    u[i] *= 0.45359237**4
                    
            xr = f(*x)
            
            if xr == 0:
                self.assertTrue(abs(gr.x - xr) < 1e-10)
            else:
                self.assertTrue(abs(gr.x - xr)/abs(xr) < 1e-10)
            
            ur = 0
            for i,y in enumerate(u):
                ur += y**2*d(i,*x)**2
            ur = np.sqrt(ur)
            if ur == 0:
                self.assertTrue(abs(gr.u) < 1e-10)
            else:
                self.assertTrue(abs(gr.u - ur)/ur < 1e-10)
            
            dofr = 0
            for i,(v,idof) in enumerate(zip(u,dof)):
                dofr += v**4*d(i,*x)**4/idof
            if dofr == 0:
                dofr = float('inf')
            else:
                dofr = ur**4/dofr
            if dofr > uc.gummy.max_dof:
                dofr = float('inf')
                
            # Due to the fact that any dof larger than gummy.mox_dof is rounded to 
            # infinity in intermediate calculations, gr.dof may not be exactly dofr
            # particularly for large dofr.  
            #if dofr > 100:
                #self.assertTrue(gr.dof > 100
            #else:
                #self.assertTrue(abs(gr.dof - dofr)/dofr < 1e-2
                
            if unt == '1':
                if k == 1:
                    dr = lambda *x: d(0,*x)
                else:
                    dr = lambda *x: [d(i,*x) for i in range(k)]
                ga = uc.gummy.apply(f,dr,*g)
                if xr == 0:
                    self.assertTrue(abs(ga.x) < 1e-10)
                else:
                    self.assertTrue(abs(ga.x - xr)/abs(xr) < 1e-10)
                if ur == 0:
                    self.assertTrue(abs(ga.u) < 1e-6)
                else:
                    self.assertTrue(abs(ga.u - ur)/ur < 1e-6)
                if dofr > 100:
                    self.assertTrue(ga.dof > 100)
                else:
                    self.assertTrue(abs(ga.dof - dofr)/dofr < 1e-2)
                    
                gn = uc.gummy.napply(f,*g)
                if xr == 0:
                    self.assertTrue(abs(gn.x) < 1e-10)
                else:
                    self.assertTrue(abs(gn.x - xr)/abs(xr) < 1e-10)
                if ur == 0:
                    self.assertTrue(abs(gn.u) < 1e-3)
                else:
                    self.assertTrue(abs(gn.u - ur)/ur < 1e-3)
                if dofr > 100:
                    self.assertTrue(gn.dof > 100)
                else:
                    self.assertTrue(abs(gn.dof - dofr)/dofr < 1e-2)
                    
                if rand.randint(2):
                    if rand.randint(2):
                        gf = ga
                    else:
                        gf = gn
                else:
                    gf = gr
            else:
                gf = gr
                
            gl = list(g)
            if not rand.randint(4):
                gl.append(make_gummy(unit=unt)[0])
                
            if rand.randint(2):
                e = gl[rand.randint(len(gl))]
            else:
                e = [gl[rand.randint(len(gl))] for i in range(rand.randint(len(gl)))]
                
            if ur > 0:
                ua = 0
                ub = 0
                ud = 0
                ue = 0
                dofa = 0
                dofb = 0
                dofd = 0
                dofe = 0
                for i,(ig,iu,idof) in enumerate(zip(g,u,dof)):
                    if gf.ufrom(ig) == 0:
                        self.assertTrue(iu*d(i,*x)/ur < 1e-3)
                    else:
                        if ig.utype == 'A':
                            ua += iu**2*d(i,*x)**2
                            dofa += iu**4*d(i,*x)**4/idof
                        elif ig.utype == 'B':
                            ub += iu**2*d(i,*x)**2
                            dofb += iu**4*d(i,*x)**4/idof
                        elif ig.utype == 'D':
                            ud += iu**2*d(i,*x)**2
                            dofd += iu**4*d(i,*x)**4/idof
                        
                        try:
                            if e is ig or ig in e:
                                ue += iu**2*d(i,*x)**2
                                dofe += iu**4*d(i,*x)**4/idof
                        except TypeError:
                            pass
                    
                ua = np.sqrt(ua)
                ub = np.sqrt(ub)
                ud = np.sqrt(ud)
                ue = np.sqrt(ue)
                
                if dofa == 0:
                    dofa = float('inf')
                else:
                    dofa  = ua**4/dofa
                if dofb == 0:
                    dofb = float('inf')
                else:
                    dofb  = ub**4/dofb
                if dofd == 0:
                    dofd = float('inf')
                else:
                    dofd  = ud**4/dofd
                if dofe == 0:
                    dofe = float('inf')
                else:
                    dofe  = ue**4/dofe
                            
                if ua == 0:
                    self.assertTrue(gf.ufrom('A') == 0)
                else:
                    self.assertTrue(abs(gf.ufrom('A') - ua)/ur < 1e-3)
                if ua/ur > 0.01:
                    if dofa > 100 or gf.ufrom('A') == 0:
                        self.assertTrue(gf.doffrom('A') > 100)
                    else:
                        self.assertTrue(abs(gf.doffrom('A') - dofa)/dofa < 1e-2)
                    
                if ub == 0:
                    self.assertTrue(gf.ufrom('B') == 0)
                else:
                    self.assertTrue(abs(gf.ufrom('B') - ub)/ur < 1e-3)
                if ub/ur > 0.01:
                    if dofb > 100 or gf.ufrom('B') == 0:
                        self.assertTrue(gf.doffrom('B') > 100)
                    else:
                        self.assertTrue(abs(gf.doffrom('B') - dofb)/dofb < 1e-2)
                    
                if ud == 0:
                    self.assertTrue(gf.ufrom('D') == 0)
                else:
                    self.assertTrue(abs(gf.ufrom('D') - ud)/ur < 1e-3)
                if ud/ur > 0.01:
                    if dofd > 100 or gf.ufrom('D') == 0:
                        self.assertTrue(gf.doffrom('D') > 100)
                    else:
                        self.assertTrue(abs(gf.doffrom('D') - dofd)/dofd < 1e-2)
                    
                if ue == 0:
                    self.assertTrue(gf.ufrom(e) == 0)
                else:
                    self.assertTrue(abs(gf.ufrom(e) - ue)/ur < 1e-3)
                #if ue/ur > 0.01:
                    #if dofe > 100 or gf.ufrom(e) == 0:
                        #self.assertTrue(gf.doffrom(e) > 100
                    #else:
                        #self.assertTrue(abs(gf.doffrom(e) - dofe)/dofe < 1e-2
                        
            if not rand.randint(4):
                gf.uunit = '%'
                    
            styles = ['pm','pmi','concise','ueq','u','uf']
            gf.style = styles[rand.randint(6)]
                    
            self.assertTrue('?' not in gf.tostring())
            
            ub = {i.utype for i in g if i.utype is not None}
            if not rand.randint(8) and len(ub) > 1:
                ub.pop()
            ub = list(ub)
            
            gfub = gf.copy(formatting=True)
            if len(ub) > 0:
                gfub.ubreakdown = ub
            
            self.assertTrue('?' not in gfub.tostring())
                    
            if prnt:
                w = rand.randint(5)
                if w == 0:
                    print(gf)
                    print(gfub)
                elif w == 1:
                    gf.latex()
                    gfub.latex()
                elif w == 2:
                    gf.html()
                    gfub.html()
                elif w == 3:
                    gf.unicode()
                    gfub.unicode()
                else:
                    gf.ascii()
                    gfub.ascii()
            
            if budget:
                show_expanded_u = bool(rand.randint(2))
                show_subtotals = bool(rand.randint(2))
                
                k = None
                p = None
                if bool(rand.randint(2)):
                    if bool(rand.randint(2)):
                        if bool(rand.randint(2)):
                            gf.k = 2
                        else:
                            k = 2
                    else:
                        if bool(rand.randint(2)):
                            gf.p = 0.95
                        else:
                            p = 0.95
                            
                if bool(rand.randint(2)):
                    gl[0].name = 'gl0'
                    
                if bool(rand.randint(2)) and len(gl) > 1:
                    gl[1].name = 'gl1'
                    
                if bool(rand.randint(2)):
                    gf.name = 'gf'
                
                if not bool(rand.randint(4)):
                    description = ['description ' + str(i) for i in range(len(gl)+1)]
                else:
                    description = None
                    
                if not bool(rand.randint(4)):
                    custom = [str(i) for i in range(len(gl)+1)]
                    custom_heading = 'n'
                else:
                    custom = None
                    custom_heading = None
                    
                show_s = None
                show_c = None
                show_d = None
                if bool(rand.randint(2)):
                    show_s = bool(rand.randint(2))
                    show_c = bool(rand.randint(2))
                    show_d = bool(rand.randint(2))
                    
                b = gf.budget(gl,show_subtotals=show_subtotals,
                              show_expanded_u=show_expanded_u,k=k,p=p,
                              description=description,show_s=show_s,show_c=show_c,
                              show_d=show_d,custom=custom,
                              custom_heading=custom_heading)
                
                w = rand.randint(5)
                if w == 0:
                    print(b)
                elif w == 1:
                    b.latex()
                elif w == 2:
                    b.html()
                else:
                    b.unicode()
            
            if prnt or budget:
                print()
                print('---')
                print()
    

if __name__ == '__main__':
    unittest.main()